# -*- coding: utf-8 -*-
import json
from typing import List

import pandas as pd

from cognite.client._api_client import APIClient


class Column:
    """Data transfer object for a column.

    Args:
        id (int):           ID of the column.
        name (str):         Name of the column.
        external_id (str):  External ID of the column.
        value_type (str):   Data type of the column.
        metadata (dict):    Custom, application specific metadata. String key -> String Value.
    """

    def __init__(
        self, id: int = None, name: str = None, external_id: str = None, value_type: str = None, metadata: dict = None
    ):
        if value_type is None:
            raise ValueError("value_type must not be None")
        self.id = id
        self.name = name
        self.externalId = external_id
        self.valueType = value_type
        self.metadata = metadata

    @staticmethod
    def from_JSON(the_column: dict):
        return Column(
            id=the_column["id"],
            name=the_column["name"],
            external_id=the_column.get("externalId", None),
            value_type=the_column["valueType"],
            metadata=the_column["metadata"],
        )


class Sequence:
    """Data transfer object for a sequence.

    Args:
        id (int):           ID of the sequence.
        name (str):         Name of the sequence.
        external_id (str):  External ID of the sequence.
        asset_id (int):     ID of the asset the sequence is connected to, if any.
        columns (List[Column]):     List of columns in the sequence.
        description (str):  Description of the sequence.
        metadata (dict):    Custom, application specific metadata. String key -> String Value.

    """

    def __init__(
        self,
        id: int = None,
        name: str = None,
        external_id: str = None,
        asset_id: int = None,
        columns: List[Column] = None,
        description: str = None,
        metadata: dict = None,
    ):
        if columns is None:
            raise ValueError("columns must not be None")
        self.id = id
        self.name = name
        self.externalId = external_id
        self.assetId = asset_id
        self.columns = columns
        self.description = description
        self.metadata = metadata

    @staticmethod
    def from_JSON(the_sequence: dict):
        return Sequence(
            id=the_sequence["id"],
            name=the_sequence.get("name", None),
            external_id=the_sequence.get("externalId", None),
            asset_id=the_sequence.get("assetId", None),
            columns=[Column.from_JSON(the_column) for the_column in the_sequence["columns"]],
            description=the_sequence["description"],
            metadata=the_sequence["metadata"],
        )


class RowValue:
    """Data transfer object for the value in a row in a sequence.

    Args:
        column_id (int):  The ID of the column that this value is for.
        value (str):      The actual value.
    """

    def __init__(self, column_id: int, value: str):
        self.columnId = column_id
        self.value = value

    @staticmethod
    def from_JSON(the_row_value: dict):
        return RowValue(column_id=the_row_value["columnId"], value=the_row_value["value"])


class Row:
    """Data transfer object for a row of data in a sequence.

    Args:
        row_number (int):  The row number for this row.
        values (list):     The values in this row.
    """

    def __init__(self, row_number: int, values: List[RowValue]):
        self.rowNumber = row_number
        self.values = values

    @staticmethod
    def from_JSON(the_row: dict):
        return Row(
            row_number=the_row["rowNumber"],
            values=[RowValue.from_JSON(the_row_value) for the_row_value in the_row["values"]],
        )

    def get_row_as_csv(self):
        return ",".join([str(x.value) for x in self.values])


class SequenceDataResponse:
    """Data transfer object for the data in a sequence, used when receiving data.

    Args:
        rows (list):  List of rows with the data.
    """

    def __init__(self, rows: List[Row]):
        self.rows = rows

    @staticmethod
    def from_JSON(the_data: dict):
        return SequenceDataResponse(rows=[Row.from_JSON(the_row) for the_row in the_data["rows"]])

    @staticmethod
    def _row_has_value_for_column(row: Row, column_id: int):
        return column_id in [value.columnId for value in row.values]

    @staticmethod
    def _get_value_for_column(row: Row, column_id: int):
        return next(value.value for value in row.values if value.columnId == column_id)

    def to_pandas(self):
        """Returns data as a pandas dataframe"""

        # Create the empty dataframe
        column_ids = [value.columnId for value in self.rows[0].values]
        my_df = pd.DataFrame(columns=column_ids)
        # Fill the dataframe with values. We might not have data for every column, so we need to be careful
        for row in self.rows:
            data_this_row = []
            for column_id in column_ids:
                # Do we have a value for this column?
                if self._row_has_value_for_column(row, column_id):
                    data_this_row.append(self._get_value_for_column(row, column_id))
                else:
                    data_this_row.append("null")
            my_df.loc[len(my_df)] = data_this_row
        return my_df

    def to_json(self):
        """Returns data as a json object"""
        raise NotImplementedError


class SequenceDataRequest:
    """Data transfer object for requesting sequence data.

    Args:
        inclusive_from (int):    Row number to get from (inclusive).
        inclusive_to (int):      Row number to get to (inclusive).
        limit (int):             How many rows to return.
        column_ids (List[int]):  ids of the columns to get data for.
    """

    def __init__(self, inclusive_from: int, inclusive_to: int, limit: int = 100, column_ids: List[int] = None):
        self.inclusiveFrom = inclusive_from
        self.inclusiveTo = inclusive_to
        self.limit = limit
        self.columnIds = column_ids or []


class SequencesClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.6", **kwargs)

    def post_sequences(self, sequences: List[Sequence]) -> Sequence:
        """Create a new time series.

        Args:
            sequences (list[test_experimental.dto.Sequence]):  List of sequence data transfer objects to create.

        Returns:
            client.test_experimental.sequences.Sequence: The created sequence
        """
        url = "/sequences"

        # Remove the id field from the sequences to be posted, as including them will lead to 400's since sequences that
        # are not created yet should not have id's yet.
        for sequence in sequences:
            del sequence.id
            for column in sequence.columns:
                del column.id

        body = {"items": [sequence.__dict__ for sequence in sequences]}
        res = self._post(url, body=body)
        json_response = json.loads(res.text)
        the_sequence = json_response["data"]["items"][0]
        return Sequence.from_JSON(the_sequence)

    def get_sequence_by_id(self, id: int) -> Sequence:
        """Returns a Sequence object containing the requested sequence.

        Args:
            id (int):       ID of the sequence to look up

        Returns:
            client.test_experimental.sequences.Sequence: A data object containing the requested sequence.
        """
        url = "/sequences/{}".format(id)
        res = self._get(url=url)
        json_response = json.loads(res.text)
        the_sequence = json_response["data"]["items"][0]
        return Sequence.from_JSON(the_sequence)

    def get_sequence_by_external_id(self, external_id: str) -> Sequence:
        """Returns a Sequence object containing the requested sequence.

        Args:
            external_id (int):  External ID of the sequence to look up

        Returns:
            test_experimental.dto.Sequence: A data object containing the requested sequence.
        """
        url = "/sequences"
        params = {"externalId": external_id}
        res = self._get(url=url, params=params)
        json_response = json.loads(res.text)
        the_sequence = json_response["data"]["items"][0]
        return Sequence.from_JSON(the_sequence)

    def list_sequences(self, external_id: str = None):
        """Returns a list of Sequence objects.

        Args:
            external_id (int, optional):  External ID of the sequence to look up

        Returns:
            List[test_experimental.dto.Sequence]: A data object containing the requested sequence.
        """
        url = "/sequences"
        params = {"externalId": external_id}
        res = self._get(url=url, params=params)
        json_response = json.loads(res.text)
        sequences = json_response["data"]["items"]
        return [Sequence.from_JSON(seq) for seq in sequences]

    def delete_sequence_by_id(self, id: int) -> None:
        """Deletes the sequence with the given id.

        Args:
            id (int):       ID of the sequence to delete

        Returns:
            None
        """
        url = "/sequences/{}".format(id)
        self._delete(url=url)

    def post_data_to_sequence(self, id: int, rows: List[Row]) -> None:
        """Posts data to a sequence.

        Args:
            id (int):       ID of the sequence.
            rows (list):    List of rows with the data.

        Returns:
            None
        """
        url = "/sequences/{}/postdata".format(id)
        body = {"items": [{"rows": [row.__dict__ for row in rows]}]}
        self._post(url, body=body)

    def get_data_from_sequence(
        self,
        id: int,
        inclusive_from: int = None,
        inclusive_to: int = None,
        limit: int = 100,
        column_ids: List[int] = None,
    ) -> SequenceDataResponse:
        """Gets data from the given sequence.

        Args:
            id (int):                id of the sequence.
            inclusive_from (int):    Row number to get from (inclusive). If set to None, you'll get data from the first row
                                     that exists.
            inclusive_to (int):      Row number to get to (inclusive). If set to None, you'll get data to the last row that
                                     exists (depending on the limit).
            limit (int):             How many rows to return.
            column_ids (List[int]):  ids of the columns to get data for.

        Returns:
            client.test_experimental.sequences.SequenceDataResponse: A data object containing the requested sequence.
        """
        url = "/sequences/{}/getdata".format(id)
        sequenceDataRequest = SequenceDataRequest(
            inclusive_from=inclusive_from, inclusive_to=inclusive_to, limit=limit, column_ids=column_ids or []
        )
        body = {"items": [sequenceDataRequest.__dict__]}
        res = self._post(url=url, body=body)
        json_response = json.loads(res.text)
        the_data = json_response["data"]["items"][0]
        return SequenceDataResponse.from_JSON(the_data)



import os
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Any, Dict, List

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResponse
from cognite.client.exceptions import APIError


class ModelResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.id = item["id"]
        self.name = item["name"]
        self.project = item["project"]
        self.description = item["description"]
        self.created_time = item["createdTime"]
        self.metadata = item["metadata"]
        self.is_deprecated = item["isDeprecated"]
        self.active_version_id = item["activeVersionId"]
        self.input_fields = item["inputFields"]
        self.output_fields = item["outputFields"]


class ModelCollectionResponse(CogniteCollectionResponse):
    _RESPONSE_CLASS = ModelResponse


class ModelVersionResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.id = item["id"]
        self.is_deprecated = item["isDeprecated"]
        self.training_details = item["trainingDetails"]
        self.name = item["name"]
        self.error_msg = item["errorMsg"]
        self.model_id = item["modelId"]
        self.created_time = item["createdTime"]
        self.metadata = item["metadata"]
        self.source_package_id = item["sourcePackageId"]
        self.status = item["status"]
        self.description = item["description"]
        self.project = item["project"]


class ModelVersionCollectionResponse(CogniteCollectionResponse):
    _RESPONSE_CLASS = ModelVersionResponse


class ModelArtifactResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.name = item["name"]
        self.size = item["size"]


class ModelArtifactCollectionResponse(CogniteCollectionResponse):
    _RESPONSE_CLASS = ModelArtifactResponse


class ModelLogResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.prediction_logs = item["predict"]
        self.training_logs = item["train"]

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]


class PredictionError(APIError):
    pass


class EmptyAritfactsDirectory(Exception):
    pass


class ModelsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.6", **kwargs)
        self._LIMIT = 1000

    def create_model(
        self,
        name: str,
        description: str = "",
        metadata: Dict[str, Any] = None,
        input_fields: List[Dict[str, str]] = None,
        output_fields: List[Dict[str, str]] = None,
    ) -> ModelResponse:
        """Creates a new model

        Args:
            name (str):             Name of model
            description (str):      Description
            metadata (Dict[str, Any]):          Metadata about model
            input_fields (List[str]):   List of input fields the model accepts
            output_fields (List[str]:   List of output fields the model produces

        Returns:
            experimental.model_hosting.models.ModelResponse: The created model.
        """
        url = "/analytics/models"
        model_body = {
            "name": name,
            "description": description,
            "metadata": metadata or {},
            "inputFields": input_fields or [],
            "outputFields": output_fields or [],
        }
        res = self._post(url, body=model_body)
        return ModelResponse(res.json())

    def list_models(self, limit: int = None, cursor: int = None, autopaging: bool = False) -> ModelCollectionResponse:
        """List all models.

        Args:
            limit (int): Maximum number of models to return. Defaults to 250.
            cursor (str): Cursor to use to fetch next set of results.
            autopaging (bool): Whether or not to automatically page through all results. Will disregard limit.

        Returns:
            experimental.model_hosting.models.ModelCollectionResponse: List of models
        """
        url = "/analytics/models"
        params = {"cursor": cursor, "limit": limit if autopaging is False else self._LIMIT}
        res = self._get(url, params=params, autopaging=autopaging)
        return ModelCollectionResponse(res.json())

    def get_model(self, id: int) -> ModelResponse:
        """Get a model by id.

        Args:
            id (int): Id of model to get.

        Returns:
            experimental.model_hosting.models.ModelResponse: The requested model
        """
        url = "/analytics/models/{}".format(id)
        res = self._get(url)
        return ModelResponse(res.json())

    def update_model(
        self, id: int, description: str = None, metadata: Dict[str, str] = None, active_version_id: int = None
    ) -> ModelResponse:
        """Update a model.

        Args:
            id (int): Id of model to update.
            description (str): Description of model.
            metadata (Dict[str, str]): metadata about model.
            active_version_id (int): Active version of model.

        Returns:
            experimental.model_hosting.models.ModelResponse: Updated model
        """
        url = "/analytics/models/{}/update".format(id)
        body = {}
        if description:
            body.update({"description": {"set": description}})
        if metadata:
            body.update({"metadata": {"set": metadata}})
        if active_version_id:
            body.update({"activeVersionId": {"set": active_version_id}})
        res = self._put(url, body=body)
        return ModelResponse(res.json())

    def deprecate_model(self, id: int) -> ModelResponse:
        """Deprecate a model.

        Args:
            id (int): Id of model to deprecate.

        Returns:
            experimental.model_hosting.models.ModelResponse: Deprecated model
        """
        url = "/analytics/models/{}/deprecate".format(id)
        res = self._put(url)
        return ModelResponse(res.json())

    def delete_model(self, id: int) -> None:
        """Delete a model.

        Will also delete all versions and schedules for this model.

        Args:
            id (int): Delete model with this id.

        Returns:
            None
        """
        url = "/analytics/models/{}".format(id)
        self._delete(url)

    def create_model_version(
        self, name: str, model_id: int, source_package_id: int, description: str = None, metadata: Dict = None
    ) -> ModelVersionResponse:
        """Create a model version without deploying it.

        Then you can optionally upload artifacts to the model version and later deploy it.

        Args:
            name (str): Name of the the model version.
            model_id (int): Create the version on the model with this id.
            source_package_id (int): Use the source package with this id. The source package must have an available
                predict operation.
            description (str):  Description of model version
            metadata (Dict[str, Any]):  Metadata about model version

        Returns:
            ModelVersionResponse: The created model version.
        """
        url = "/analytics/models/{}/versions".format(model_id)
        body = {
            "name": name,
            "description": description or "",
            "sourcePackageId": source_package_id,
            "metadata": metadata or {},
        }
        res = self._post(url, body=body)
        return ModelVersionResponse(res.json())

    def deploy_awaiting_model_version(self, model_id: int, version_id: int) -> ModelVersionResponse:
        """Deploy an already created model version awaiting manual deployment.

        The model version must have status AWAITING_MANUAL_DEPLOYMENT in order for this to work.

        Args:
            model_id (int): The id of the model containing the version to deploy.
            version_id (int): The id of the model version to deploy.
        Returns:
            ModelVersionResponse: The deployed model version.
        """
        url = "/analytics/models/{}/versions/{}/deploy".format(model_id, version_id)
        res = self._post(url, body={})
        return ModelVersionResponse(res.json())

    def deploy_model_version(
        self,
        name: str,
        model_id: int,
        source_package_id: int,
        artifacts_directory: str = None,
        description: str = None,
        metadata: Dict = None,
    ) -> ModelVersionResponse:
        """This will create and deploy a model version.

        If artifacts_directory is specified, it will traverse that directory recursively and
        upload all artifacts in that directory before deploying.

        Args:
            name (str): Name of the the model version.
            model_id (int): Create the version on the model with this id.
            source_package_id (int): Use the source package with this id. The source package must have an available
                predict operation.
            artifacts_directory (str, optional): Absolute path of directory containing artifacts.
            description (str, optional):  Description of model version
            metadata (Dict[str, Any], optional):  Metadata about model version
            
        Returns:
            ModelVersionResponse: The created model version.
        """
        model_version = self.create_model_version(name, model_id, source_package_id, description, metadata)
        if artifacts_directory:
            self.upload_artifacts_from_directory(model_id, model_version.id, directory=artifacts_directory)
        return self.deploy_awaiting_model_version(model_id, model_version.id)

    def train_and_deploy_model_version(
        self,
        name: str,
        model_id: int,
        source_package_id: int,
        train_source_package_id: int = None,
        metadata: Dict = None,
        description: str = None,
        args: Dict[str, Any] = None,
        scale_tier: str = None,
        machine_type: str = None,
    ) -> ModelVersionResponse:
        """Train and deploy a new version of a model.

        This will instantiate a training job and automatically deploy the model upon completion.

        Args:
            model_id (int): Create a new version under the model with this id
            name (str): Name of model version.
            source_package_id (int):    Use the source package with this id
            train_source_package_id (int):  Use this source package for training. If omitted, will default to
                                            source_package_id.
            metadata (Dict[str, Any]):  Metadata about model version
            description (str):  Description of model version
            args (Dict[str, Any]):   Dictionary of arguments to pass to the training job.
            scale_tier (str):   Which scale tier to use. Must be either "BASIC" or "CUSTOM".
            machine_type (str): Specify a machine type. Applies only if scale_tier is "CUSTOM".

        Returns:
            experimental.model_hosting.models.ModelVersionResponse: The created model version.
        """
        url = "/analytics/models/{}/versions/train".format(model_id)
        if args and "data_spec" in args:
            data_spec = args["data_spec"]
            if hasattr(data_spec, "dump"):
                args["data_spec"] = data_spec.dump()
        body = {
            "name": name,
            "description": description or "",
            "sourcePackageId": source_package_id,
            "trainingDetails": {
                "sourcePackageId": train_source_package_id or source_package_id,
                "args": args or {},
                "scaleTier": scale_tier or "BASIC",
                "machineType": machine_type,
            },
            "metadata": metadata or {},
        }
        res = self._post(url, body=body)
        return ModelVersionResponse(res.json())

    def list_model_versions(
        self, model_id: int, limit: int = None, cursor: str = None, autopaging: bool = False
    ) -> ModelVersionCollectionResponse:
        """Get all versions of a specific model.

        Args:
            model_id (int): Get versions for the model with this id.
            limit (int): Maximum number of model versions to return. Defaults to 250.
            cursor (str): Cursor to use to fetch next set of results.
            autopaging (bool): Whether or not to automatically page through all results. Will disregard limit.

        Returns:
            experimental.model_hosting.models.ModelVersionCollectionResponse: List of model versions
        """
        url = "/analytics/models/{}/versions".format(model_id)
        params = {"cursor": cursor, "limit": limit if autopaging is False else self._LIMIT}
        res = self._get(url, params=params, autopaging=True)
        return ModelVersionCollectionResponse(res.json())

    def get_model_version(self, model_id: int, version_id: int) -> ModelVersionResponse:
        """Get a specific model version by id.

        Args:
            model_id (int): Id of model which has the model version.
            version_id (int): Id of model version.

        Returns:
            experimental.model_hosting.models.ModelVersionResponse: The requested model version
        """
        url = "/analytics/models/{}/versions/{}".format(model_id, version_id)
        res = self._get(url)
        return ModelVersionResponse(res.json())

    def update_model_version(
        self, model_id: int, version_id: int, description: str = None, metadata: Dict[str, str] = None
    ) -> ModelVersionResponse:
        """Update description or metadata on a model version.

        Args:
            model_id (int): Id of model containing the model version.
            version_id (int): Id of model version to update.
            description (str): New description.
            metadata (Dict[str, str]): New metadata

        Returns:
            ModelVersionResponse: The updated model version.
        """
        url = "/analytics/models/{}/versions/{}/update".format(model_id, version_id)
        body = {}
        if description:
            body.update({"description": {"set": description}})
        if metadata:
            body.update({"metadata": {"set": metadata}})

        res = self._put(url, body=body)
        return ModelVersionResponse(res.json())

    def deprecate_model_version(self, model_id: int, version_id: int) -> ModelVersionResponse:
        """Deprecate a model version

        Args:
            model_id (int): Id of model
            version_id (int): Id of model version to deprecate

        Returns:
            ModelVersionResponse: The deprecated model version
        """
        url = "/analytics/models/{}/versions/{}/deprecate".format(model_id, version_id)
        res = self._put(url)
        return ModelVersionResponse(res.json())

    def delete_model_version(self, model_id: int, version_id: int) -> None:
        """Delete a model version by id.

        Args:
            model_id (int): Id of model which has the model version.
            version_id (int): Id of model version.

        Returns:
            None
        """
        url = "/analytics/models/{}/versions/{}".format(model_id, version_id)
        self._delete(url)

    def online_predict(
        self, model_id: int, version_id: int = None, instances: List = None, args: Dict[str, Any] = None
    ) -> List:
        """Perform online prediction on a models active version or a specified version.

        Args:
            model_id (int):     Perform a prediction on the model with this id. Will use active version.
            version_id (int):   Use this version instead of the active version. (optional)
            instances (List): List of JSON serializable instances to pass to your model one-by-one.
            args (Dict[str, Any])    Dictinoary of keyword arguments to pass to your predict method.

        Returns:
            List: List of predictions for each instance.
        """
        url = "/analytics/models/{}/predict".format(model_id)
        if instances:
            for i, instance in enumerate(instances):
                if hasattr(instance, "dump"):
                    instances[i] = instance.dump()
        if version_id:
            url = "/analytics/models/{}/versions/{}/predict".format(model_id, version_id)
        body = {"instances": instances, "args": args or {}}
        res = self._put(url, body=body).json()
        if "error" in res:
            raise PredictionError(message=res["error"]["message"], code=res["error"]["code"])
        return res["data"]["predictions"]

    def list_artifacts(self, model_id: int, version_id: int) -> ModelArtifactCollectionResponse:
        """List the artifacts associated with the specified model version.

        Args:
            model_id (int): Id of model
            version_id: Id of model version to get artifacts for

        Returns:
            experimental.model_hosting.models.ModelArtifactCollectionResponse: List of artifacts
        """
        url = "/analytics/models/{}/versions/{}/artifacts".format(model_id, version_id)
        res = self._get(url)
        return ModelArtifactCollectionResponse(res.json())

    def download_artifact(self, model_id: int, version_id: int, name: str, directory: str = None) -> None:
        """Download an artifact to a directory. Defaults to current working directory.

        Args:
            model_id (int): Id of model
            version_id (int): Id of model version.
            name (int): Name of artifact.
            directory (int): Directory to place artifact in. Defaults to current working directory.

        Returns:
            None
        """
        directory = directory or os.getcwd()
        file_path = os.path.join(directory, name)

        url = "/analytics/models/{}/versions/{}/artifacts/{}".format(model_id, version_id, name)
        download_url = self._get(url).json()["data"]["downloadUrl"]
        with open(file_path, "wb") as fh:
            response = self._request_session.get(download_url).content
            fh.write(response)

    def upload_artifact_from_file(self, model_id: int, version_id: int, name: str, file_path: str) -> None:
        """Upload an artifact to a model version.

        The model version must have status AWAITING_MANUAL_DEPLOYMENT in order for this to work.

        Args:
            model_id (int): The id of the model.
            version_id (int): The id of the model version to upload the artifacts to.
            name (str): The name of the artifact.
            file_path (str): The local path of the artifact.
        Returns:
            None
        """
        url = "/analytics/models/{}/versions/{}/artifacts/upload".format(model_id, version_id)
        body = {"name": name}
        res = self._post(url, body=body)
        upload_url = res.json()["data"]["uploadUrl"]
        self._upload_file(upload_url, file_path)

    def upload_artifacts_from_directory(self, model_id: int, version_id: int, directory: str) -> None:
        """Upload all files in directory recursively.

        Args:
            model_id (int): The id of the model.
            version_id (int): The id of the model version to upload the artifacts to.
            directory (int): Absolute path of directory to upload artifacts from.
        Returns:
            None
        """
        upload_tasks = []
        for root, dirs, files in os.walk(directory):
            for file_name in files:
                file_path = os.path.join(root, file_name)
                full_file_name = os.path.relpath(file_path, directory)
                upload_tasks.append((model_id, version_id, full_file_name, file_path))

        if len(upload_tasks) == 0:
            raise EmptyAritfactsDirectory("Artifacts directory is empty.")
        self._execute_tasks_concurrently(self.upload_artifact_from_file, upload_tasks)

    @staticmethod
    def _execute_tasks_concurrently(func, tasks):
        with ThreadPoolExecutor(16) as p:
            futures = [p.submit(func, *task) for task in tasks]
            return [future.result() for future in futures]

    def _upload_file(self, upload_url, file_path):
        with open(file_path, "rb") as fh:
            mydata = fh.read()
            response = self._request_session.put(upload_url, data=mydata)
        return response

    def get_logs(self, model_id: int, version_id: int, log_type: str = None) -> ModelLogResponse:
        """Get logs for prediction and/or training routine of a specific model version.

        Args:
            model_id (int): Id of model.
            version_id (int): Id of model version to get logs for.
            log_type (str): Which routine to get logs from. Must be 'train’, 'predict’, or ‘both’. Defaults to 'both'.

        Returns:
            ModelLogResponse: An object containing the requested logs.
        """
        url = "/analytics/models/{}/versions/{}/log".format(model_id, version_id)
        params = {"logType": log_type}
        res = self._get(url, params=params)
        return ModelLogResponse(res.json())



from cognite.client.experimental.model_hosting.models import ModelsClient
from cognite.client.experimental.model_hosting.schedules import SchedulesClient
from cognite.client.experimental.model_hosting.source_packages import SourcePackageClient


class ModelHostingClient:
    def __init__(self, client_factory):
        self._client_factory = client_factory
        self._models_client = self._client_factory(ModelsClient)
        self._source_package_client = self._client_factory(SourcePackageClient)
        self._schedules_client = self._client_factory(SchedulesClient)

    @property
    def models(self) -> ModelsClient:
        return self._models_client

    @property
    def source_packages(self) -> SourcePackageClient:
        return self._source_package_client

    @property
    def schedules(self) -> SchedulesClient:
        return self._schedules_client



# -*- coding: utf-8 -*-
from copy import copy
from typing import List

import pandas as pd

from cognite.client._api_client import APIClient, CogniteResponse


class TimeSeriesResponse(CogniteResponse):
    """Time series Response Object"""

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"]

    def to_pandas(self, include_metadata: bool = False):
        """Returns data as a pandas dataframe

        Args:
            include_metadata (bool): Whether or not to include metadata fields in the resulting dataframe
        """
        items = copy(self.internal_representation["data"]["items"])
        if items and items[0].get("metadata") is None:
            return pd.DataFrame(items)
        for d in items:
            if d.get("metadata"):
                metadata = d.pop("metadata")
                if include_metadata:
                    d.update(metadata)
        return pd.DataFrame(items)


class TimeSeriesClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.6", **kwargs)

    def delete_time_series_by_id(self, ids: List[int]) -> None:
        """Delete multiple time series by id.

        Args:
            ids (List[int]):   IDs of time series to delete.

        Returns:
            None

        Examples:
            Delete a single time series by id::

                client = CogniteClient()

                client.time_series.delete_time_series_by_id(ids=[my_ts_id])
        """
        url = "/timeseries/delete"
        body = {"items": ids}
        self._post(url, body=body)

    def get_time_series_by_id(self, id: int) -> TimeSeriesResponse:
        """Returns a TimeseriesResponse object containing the requested timeseries.

        Args:
            id (int):           ID of timeseries to look up

        Returns:
            client.experimental.time_series.TimeSeriesResponse: A data object containing the requested timeseries.
        """
        url = "/timeseries/{}".format(id)
        params = {}
        res = self._get(url=url, params=params)
        return TimeSeriesResponse(res.json())

    def get_multiple_time_series_by_id(self, ids: List[int]) -> TimeSeriesResponse:
        """Returns a TimeseriesResponse object containing the requested timeseries.

        Args:
            ids (List[int]):           IDs of timeseries to look up

        Returns:
            client.experimental.time_series.TimeSeriesResponse: A data object containing the requested timeseries with several
            getter methods with different output formats.
        """
        url = "/timeseries/byids"
        body = {"items": ids}
        params = {}
        res = self._post(url=url, body=body, params=params)
        return TimeSeriesResponse(res.json())

    def search_for_time_series(
        self,
        name=None,
        description=None,
        query=None,
        unit=None,
        is_string=None,
        is_step=None,
        metadata=None,
        asset_ids=None,
        asset_subtrees=None,
        min_created_time=None,
        max_created_time=None,
        min_last_updated_time=None,
        max_last_updated_time=None,
        **kwargs
    ) -> TimeSeriesResponse:
        """Returns a TimeSeriesResponse object containing the search results.

        Args:
            name (str): Prefix and fuzzy search on name.
            description (str):  Prefix and fuzzy search on description.
            query (str):    Search on name and description using wildcard search on each of the words (separated by spaces).
                            Retrieves results where at least on word must match. Example: "some other"
            unit (str): Filter on unit (case-sensitive)
            is_string (bool): Filter on whether the ts is a string ts or not.
            is_step (bool): Filter on whether the ts is a step ts or not.
            metadata (Dict):    Filter out time series that do not match these metadata fields and values (case-sensitive).
                                Format is {"key1": "val1", "key2", "val2"}
            asset_ids (List): Filter out time series that are not linked to any of these assets. Format is [12,345,6,7890].
            asset_subtrees (List):  Filter out time series that are not linked to assets in the subtree rooted at these assets.
                                    Format is [12,345,6,7890].
            min_created_time (int):   Filter out time series with createdTime before this. Format is milliseconds since epoch.
            max_created_time (int):   Filter out time series with createdTime after this. Format is milliseconds since epoch.
            min_last_updated_time (int): Filter out time series with lastUpdatedTime before this. Format is milliseconds since epoch.
            max_last_updated_time (int): Filter out time series with lastUpdatedTime after this. Format is milliseconds since epoch.

        Keyword Arguments:
            sort (str):     "createdTime" or "lastUpdatedTime". Field to be sorted.
                            If not specified, results are sorted by relevance score.
            dir (str):      "asc" or "desc". Only applicable if sort is specified. Default 'desc'.
            limit (int):    Return up to this many results. Maximum is 1000. Default is 25.
            offset (int):   Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.
            boost_name (bool): Whether or not boosting name field. This option is test_experimental and can be changed.

        Returns:
            client.experimental.time_series.TimeSeriesResponse: A data object containing the requested timeseries with several getter methods with different
            output formats.
        """
        url = "/timeseries/search"
        params = {
            "name": name,
            "description": description,
            "query": query,
            "unit": unit,
            "isString": is_string,
            "isStep": is_step,
            "metadata": str(metadata) if metadata is not None else None,
            "assetIds": str(asset_ids) if asset_ids is not None else None,
            "assetSubtrees": str(asset_subtrees) if asset_subtrees is not None else None,
            "minCreatedTime": min_created_time,
            "maxCreatedTime": max_created_time,
            "minLastUpdatedTime": min_last_updated_time,
            "maxLastUpdatedTime": max_last_updated_time,
            "sort": kwargs.get("sort"),
            "dir": kwargs.get("dir"),
            "limit": kwargs.get("limit", self._LIMIT),
            "offset": kwargs.get("offset"),
            "boostName": kwargs.get("boost_name"),
        }
        res = self._get(url, params=params)
        return TimeSeriesResponse(res.json())



# -*- coding: utf-8 -*-
import os
import warnings
from copy import copy
from typing import Dict, List, Union

import pandas as pd

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResponse


class FileInfoResponse(CogniteResponse):
    """File Info Response Object.

    Args:
        id (int):               ID given by the API to the file.
        file_name (str):        File name. Max length is 256.
        directory (str):        Directory containing the file. Max length is 512.
        source (dict):          Source that this file comes from. Max length is 256.
        file_type (str):        File type. E.g. pdf, css, spreadsheet, .. Max length is 64.
        metadata (dict):        Customized data about the file.
        asset_ids (list[str]):  Names of assets related to this file.
        uploaded (bool):        Whether or not the file is uploaded.
        uploaded_at (int):      Epoc thime (ms) when the file was uploaded succesfully.
    """

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.internal_representation["data"]["items"][0]
        self.id = item.get("id")
        self.file_name = item.get("fileName")
        self.directory = item.get("directory")
        self.source = item.get("source")
        self.file_type = item.get("fileType")
        self.metadata = item.get("metadata")
        self.asset_ids = item.get("assetIds")
        self.uploaded = item.get("uploaded")
        self.uploaded_at = item.get("uploadedAt")

    def to_pandas(self):
        file_info = copy(self.to_json())
        if file_info.get("metadata"):
            file_info.update(file_info.pop("metadata"))
        return pd.DataFrame.from_dict(file_info, orient="index")


class FileListResponse(CogniteCollectionResponse):
    """File List Response Object"""

    _RESPONSE_CLASS = FileInfoResponse

    def to_pandas(self):
        return pd.DataFrame(self.to_json())


class FilesClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def upload_file(
        self, file_name, file_path=None, directory=None, source=None, file_type=None, content_type=None, **kwargs
    ) -> Dict:
        """Upload metadata about a file and get an upload link.

        The link will expire after 30 seconds if not resumable. A resumable upload link is default. Such a link is one-time
        use and expires after one week. For more information, check this link:
        https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload. Use PUT request to upload file with the
        link returned.

        If file_path is specified, the file will be uploaded directly by the SDK.

        Args:
            file_name (str):      File name. Max length is 256.

            file_path (str, optional):     Path of file to upload, if omitted a upload link will be returned.

            content_type (str, optional):   MIME type of your file. Required if file_path is specified.

            directory (str, optional):      Directory containing the file. Max length is 512.

            source (str, optional):         Source that this file comes from. Max length is 256.

            file_type (str, optional):      File type. E.g. pdf, css, spreadsheet, .. Max length is 64.

        Keyword Args:
            metadata (dict):      Customized data about the file.

            asset_ids (list):       IDs of assets related to this file.

            resumable (bool):     Whether to generate a resumable URL or not. Default is true.

            overwrite (bool):     Whether to overwrite existing data if duplicate or not. Default is false.

        Returns:
            Dict: A dictionary containing the field fileId and optionally also uploadURL if file_path is omitted.

        Examples:
            Upload a file and link it to an asset::

                client = CogniteClient()
                res = client.files.upload_file(file_name="myfile", file_path="/path/to/my/file.txt",
                        content_type="text/plain", asset_ids=[123])
                file_id = res["fileId"]
        """
        url = "/files/initupload"

        headers = {"X-Upload-Content-Type": content_type}

        params = {"resumable": kwargs.get("resumable", True), "overwrite": kwargs.get("overwrite", False)}

        body = {
            "fileName": file_name,
            "directory": directory,
            "source": source,
            "fileType": file_type,
            "metadata": kwargs.get("metadata", None),
            "assetIds": kwargs.get("asset_ids", None),
        }
        res_storage = self._post(url=url, body=body, headers=headers, params=params)
        result = res_storage.json()["data"]
        if file_path:
            if not content_type:
                warning = "content_type should be specified when directly uploading the file."
                warnings.warn(warning)
            headers = {"content-length": str(os.path.getsize(file_path))}
            with open(file_path, "rb") as file:
                self._request_session.put(result["uploadURL"], data=file, headers=headers)
            result.pop("uploadURL")
        return result

    def download_file(self, id: int, get_contents: bool = False) -> Union[str, bytes]:
        """Get list of files matching query.

        Args:
            id (int):                           Path to file to upload, if omitted a upload link will be returned.

            get_contents (bool, optional):      Boolean to determince whether or not to return file contents as string.
                                                Default is False and download url is returned.

        Returns:
            Union[str, bytes]: Download link if get_contents is False else file contents.

        Examples:
            Get a download url for the file::

                client = CogniteClient()
                res = client.files.download_file(id=12345)
                download_url = res["downloadUrl"]

            Download a file::

                client = CogniteClient()
                file_bytes = client.files.download_file(id=12345, get_contents=True)

        """
        url = "/files/{}/downloadlink".format(id)
        res = self._get(url=url)
        if get_contents:
            dl_link = res.json()["data"]
            res = self._request_session.get(dl_link)
            return res.content
        return res.json()["data"]

    def delete_files(self, file_ids) -> List:
        """Delete

        Args:
            file_ids (list[int]):   List of IDs of files to delete.

        Returns:
            List of files deleted and files that failed to delete.

        Examples:
            Delete two files::

                client = CogniteClient()
                res = client.files.delete_files([123, 234])
        """
        url = "/files/delete"
        body = {"items": file_ids}
        res = self._post(url, body=body)
        return res.json()["data"]

    def list_files(self, name=None, directory=None, file_type=None, source=None, **kwargs) -> FileListResponse:
        """Get list of files matching query.

        Args:
            name (str, optional):      List all files with this name.

            directory (str, optional):      Directory to list files from.

            source (str, optional):         List files coming from this source.

            file_type (str, optional):      Type of files to list.

        Keyword Args:
            asset_id (list):                Returns all files associated with this asset id.

            sort (str):                     Sort descending or ascending. 'ASC' or 'DESC'.

            limit (int):                    Number of results to return.

            is_uploaded (bool):             List only uploaded files if true. If false, list only other files. If not set,
                                            list all files without considering whether they are uploaded or not.

            autopaging (bool):              Whether or not to automatically page through results. If set to true, limit will
                                            be disregarded. Defaults to False.

            cursor (str):                   Cursor to use for paging through results.

        Returns:
            stable.files.FileListResponse: A data object containing the requested files information.
        Examples:
            List all files in a given directory::

                client = CogniteClient()
                res = client.files.list_files(directory="allfiles/myspecialfiles", autopaging=True)
                print(res.to_pandas())
        """
        autopaging = kwargs.get("autopaging", False)
        url = "/files"
        params = {
            "assetId": kwargs.get("asset_id"),
            "dir": directory,
            "name": name,
            "type": file_type,
            "source": source,
            "isUploaded": kwargs.get("is_uploaded"),
            "sort": kwargs.get("sort"),
            "limit": kwargs.get("limit", self._LIMIT) if not autopaging else self._LIMIT,
            "cursor": kwargs.get("cursor"),
        }

        res = self._get(url=url, params=params, autopaging=autopaging)
        return FileListResponse(res.json())

    def get_file_info(self, id) -> FileInfoResponse:
        """Returns information about a file.

        Args:
            id (int):                   Id of the file.

        Returns:
            stable.files.FileInfoResponse: A data object containing the requested file information.

        Examples:
            Get info a bout a specific file::

                client = CogniteClient()
                res = client.files.get_file_info(12345)
                print(res)
        """
        url = "/files/{}".format(id)
        res = self._get(url)
        return FileInfoResponse(res.json())



# -*- coding: utf-8 -*-
from concurrent.futures import ThreadPoolExecutor as Pool
from functools import partial
from typing import List

import pandas as pd

from cognite.client import _utils
from cognite.client._api_client import APIClient, CogniteResponse


class DatapointsResponse(CogniteResponse):
    """Datapoints Response Object."""

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"][0]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        return pd.DataFrame(self.internal_representation["data"]["items"][0]["datapoints"])


class DatapointsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.6", **kwargs)
        self._LIMIT_AGG = 10000
        self._LIMIT = 100000

    def get_datapoints(self, id, start, end=None, aggregates=None, granularity=None, **kwargs) -> DatapointsResponse:
        """Returns a DatapointsObject containing a list of datapoints for the given query.

        This method will automate paging for the user and return all data for the given time period.

        Args:
            id (int):             The unique id of the timeseries to retrieve data for.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        Keyword Arguments:
            include_outside_points (bool):      No description

            workers (int):        Number of download processes to run in parallell. Defaults to number returned by cpu_count().

            limit (str):            Max number of datapoints to return. If limit is specified, this method will not automate
                                    paging and will return a maximum of 100,000 dps.

        Returns:
            client.test_experimental.datapoints.DatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        start, end = _utils.interval_to_ms(start, end)

        if kwargs.get("limit"):
            return self._get_datapoints_user_defined_limit(
                id,
                aggregates,
                granularity,
                start,
                end,
                limit=kwargs.get("limit"),
                include_outside_points=kwargs.get("include_outside_points", False),
            )

        num_of_workers = kwargs.get("processes", self._num_of_workers)
        if kwargs.get("include_outside_points") is True:
            num_of_workers = 1

        windows = _utils.get_datapoints_windows(start, end, granularity, num_of_workers)

        partial_get_dps = partial(
            self._get_datapoints_helper_wrapper,
            id=id,
            aggregates=aggregates,
            granularity=granularity,
            include_outside_points=kwargs.get("include_outside_points", False),
        )

        with Pool(len(windows)) as p:
            datapoints = p.map(partial_get_dps, windows)

        concat_dps = []
        [concat_dps.extend(el) for el in datapoints]

        return DatapointsResponse({"data": {"items": [{"id": id, "datapoints": concat_dps}]}})

    def _get_datapoints_helper_wrapper(self, args, id, aggregates, granularity, include_outside_points):
        return self._get_datapoints_helper(
            id, aggregates, granularity, args["start"], args["end"], include_outside_points=include_outside_points
        )

    def _get_datapoints_helper(self, id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
        """Returns a list of datapoints for the given query.

        This method will automate paging for the given time period.

        Args:
            id (int):       The unique id of the timeseries to retrieve data for.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

        Keyword Arguments:
            include_outside_points (bool):  No description.

        Returns:
            list of datapoints: A list containing datapoint dicts.
        """
        url = "/timeseries/{}/data".format(id)

        limit = self._LIMIT if aggregates is None else self._LIMIT_AGG

        params = {
            "aggregates": aggregates,
            "granularity": granularity,
            "limit": limit,
            "start": start,
            "end": end,
            "includeOutsidePoints": kwargs.get("include_outside_points", False),
        }

        datapoints = []
        while (not datapoints or len(datapoints[-1]) == limit) and params["end"] > params["start"]:
            res = self._get(url, params=params)
            res = res.json()["data"]["items"][0]["datapoints"]

            if not res:
                break

            datapoints.append(res)
            latest_timestamp = int(datapoints[-1][-1]["timestamp"])
            params["start"] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
        dps = []
        [dps.extend(el) for el in datapoints]
        return dps

    def _get_datapoints_user_defined_limit(
        self, id: int, aggregates: List, granularity: str, start, end, limit, **kwargs
    ) -> DatapointsResponse:
        """Returns a DatapointsResponse object with the requested data.

        No paging or parallelizing is done.

        Args:
            id (int):       The unique id of the timeseries to retrieve data for.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            limit (str):            Max number of datapoints to return. Max is 100,000.

        Keyword Arguments:
            include_outside_points (bool):  No description.

        Returns:
            client.test_experimental.datapoints.DatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/timeseries/{}/data".format(id)

        params = {
            "aggregates": aggregates,
            "granularity": granularity,
            "limit": limit,
            "start": start,
            "end": end,
            "includeOutsidePoints": kwargs.get("include_outside_points", False),
        }
        res = self._get(url, params=params)
        res = res.json()["data"]["items"][0]["datapoints"]
        return DatapointsResponse({"data": {"items": [{"id": id, "datapoints": res}]}})



from typing import Any, Dict, List, Union

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResponse


class ScheduleResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.is_deprecated = item["isDeprecated"]
        self.name = item["name"]
        self.schedule_data_spec = item["dataSpec"]
        self.model_id = item["modelId"]
        self.created_time = item["createdTime"]
        self.metadata = item["metadata"]
        self.id = item["id"]
        self.args = item["args"]
        self.description = item["description"]


class ScheduleCollectionResponse(CogniteCollectionResponse):
    _RESPONSE_CLASS = ScheduleResponse


class SchedulesClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.6", **kwargs)
        self._LIMIT = 1000

    def create_schedule(
        self,
        model_id: int,
        name: str,
        schedule_data_spec: Any,
        description: str = None,
        args: Dict = None,
        metadata: Dict = None,
    ) -> ScheduleResponse:
        """Create a new schedule on a given model.

        Args:
            model_id (int): Id of model to create schedule on
            name (str): Name of schedule
            schedule_data_spec (Any): Specification of schedule input/output. Can be either a dictionary or a
                                    ScheduleDataSpec object from cognite-data-fetcher
            description (str): Description for schedule
            args (Dict): Dictionary of keyword arguments to pass to predict method.
            metadata (Dict): Dictionary of metadata about schedule

        Returns:
            experimental.model_hosting.schedules.ScheduleResponse: The created schedule.
        """
        url = "/analytics/models/schedules"

        if hasattr(schedule_data_spec, "dump"):
            schedule_data_spec = schedule_data_spec.dump()

        body = {
            "name": name,
            "description": description or "",
            "modelId": model_id,
            "args": args or {},
            "dataSpec": schedule_data_spec,
            "metadata": metadata or {},
        }
        res = self._post(url, body=body)
        return ScheduleResponse(res.json())

    def list_schedules(
        self, limit: int = None, cursor: int = None, autopaging: bool = False
    ) -> ScheduleCollectionResponse:
        """Get all schedules.

        Args:
            limit (int): Maximum number of schedules to return. Defaults to 250.
            cursor (str): Cursor to use to fetch next set of results.
            autopaging (bool): Whether or not to automatically page through all results. Will disregard limit.

        Returns:
            experimental.model_hosting.schedules.ScheduleCollectionResponse: The requested schedules.
        """
        url = "/analytics/models/schedules"
        params = {"cursor": cursor, "limit": limit if autopaging is False else self._LIMIT}
        res = self._get(url, params=params, autopaging=autopaging)
        return ScheduleCollectionResponse(res.json())

    def get_schedule(self, id: int) -> ScheduleResponse:
        """Get a schedule by id.

        Args:
            id (int): Id of schedule to get.
        Returns:
            experimental.model_hosting.schedules.ScheduleResponse: The requested schedule.
        """
        url = "/analytics/models/schedules/{}".format(id)
        res = self._get(url=url)
        return ScheduleResponse(res.json())

    def deprecate_schedule(self, id: int) -> ScheduleResponse:
        """Deprecate a schedule.

        Args:
            id (int): Id of schedule to deprecate

        Returns:
            experimental.model_hosting.schedules.ScheduleResponse: The deprecated schedule.
        """
        url = "/analytics/models/schedules/{}/deprecate".format(id)
        res = self._put(url)
        return ScheduleResponse(res.json())

    def delete_schedule(self, id: int) -> None:
        """Delete a schedule by id.

        Args:
            id (int):  The id of the schedule to delete.

        Returns:
            None
        """
        url = "/analytics/models/schedules/{}".format(id)
        self._delete(url=url)



# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import List
from urllib.parse import quote

import pandas as pd

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResource, CogniteResponse


class TimeSeriesResponse(CogniteResponse):
    """Time series Response Object"""

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.id = item.get("id")
        self.name = item.get("name")
        self.unit = item.get("unit")
        self.is_step = item.get("isStep")
        self.is_string = item.get("isString")
        self.created_time = item.get("createdTime")
        self.last_updated_time = item.get("lastUpdatedTime")
        self.metadata = item.get("metadata")
        self.asset_id = item.get("assetId")
        self.description = item.get("description")

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        if len(self.to_json()) > 0:
            ts = self.to_json().copy()
            if "metadata" in ts:
                # Hack to avoid path ending up as first element in dict as from_dict will fail
                metadata = ts.pop("metadata")
                df = pd.DataFrame.from_dict(ts, orient="index")
                df.loc["metadata"] = [metadata]
            else:
                df = pd.DataFrame.from_dict(ts, orient="index")
            return df
        return pd.DataFrame()


class TimeSeriesListResponse(CogniteCollectionResponse):
    """Time series Response Object"""

    _RESPONSE_CLASS = TimeSeriesResponse

    def to_pandas(self, include_metadata: bool = True):
        """Returns data as a pandas dataframe

        Args:
            include_metadata (bool): Whether or not to include metadata fields in the resulting dataframe
        """
        items = deepcopy(self.internal_representation["data"]["items"])
        if items and items[0].get("metadata") is None:
            return pd.DataFrame(items)
        for d in items:
            if d.get("metadata"):
                metadata = d.pop("metadata")
                if include_metadata:
                    d.update(metadata)
        return pd.DataFrame(items)


class TimeSeries(CogniteResource):
    """Data Transfer Object for a time series.

    Args:
        name (str):       Unique name of time series.
        is_string (bool):    Whether the time series is string valued or not.
        metadata (dict):    Metadata.
        unit (str):         Physical unit of the time series.
        asset_id (int):     Asset that this time series belongs to.
        description (str):  Description of the time series.
        security_categories (list(int)): Security categories required in order to access this time series.
        is_step (bool):        Whether or not the time series is a step series.

    """

    def __init__(
        self,
        name,
        is_string=False,
        metadata=None,
        unit=None,
        asset_id=None,
        description=None,
        security_categories=None,
        is_step=None,
    ):
        self.name = name
        self.is_string = is_string
        self.metadata = metadata
        self.unit = unit
        self.asset_id = asset_id
        self.description = description
        self.security_categories = security_categories
        self.is_step = is_step


class TimeSeriesClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def get_time_series(
        self, prefix=None, description=None, include_metadata=False, asset_id=None, path=None, **kwargs
    ) -> TimeSeriesListResponse:
        """Returns an object containing the requested timeseries.

        Args:
            prefix (str):           List timeseries with this prefix in the name.

            description (str):      Filter timeseries taht contains this string in its description.

            include_metadata (bool):    Decide if the metadata field should be returned or not. Defaults to False.

            asset_id (int):        Get timeseries related to this asset.

            path (List[int]):             Get timeseries under this asset path branch.

        Keyword Arguments:
            limit (int):            Number of results to return.

            autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be
                                    disregarded. Defaults to False.

        Returns:
            stable.time_series.TimeSeriesListResponse: A data object containing the requested timeseries with several getter methods with different
            output formats.

        Examples:
            Get all time series for a given asset::

                client = CogniteClient()
                res = client.time_series.get_time_series(asset_id=123, autopaging=True)
                print(res.to_pandas())
        """
        autopaging = kwargs.get("autopaging", False)
        url = "/timeseries"
        params = {
            "q": prefix,
            "description": description,
            "includeMetadata": include_metadata,
            "assetId": asset_id,
            "path": str(path) if path else None,
            "limit": kwargs.get("limit", self._LIMIT) if not autopaging else self._LIMIT,
        }

        res = self._get(url=url, params=params, autopaging=autopaging)
        return TimeSeriesListResponse(res.json())

    def post_time_series(self, time_series: List[TimeSeries]) -> None:
        """Create a new time series.

        Args:
            time_series (list[stable.time_series.TimeSeries]):   List of time series data transfer objects to create.

        Returns:
            None

        Examples:
            Create a new time series::

                from cognite.client.stable.time_series import TimeSeries
                client = CogniteClient()

                my_time_series = [TimeSeries(name="my_ts_1")]

                client.time_series.post_time_series(my_time_series)
        """
        url = "/timeseries"
        items = [ts.camel_case_dict() for ts in time_series]
        body = {"items": items}
        self._post(url, body=body)

    def update_time_series(self, time_series: List[TimeSeries]) -> None:
        """Update an existing time series.

        For each field that can be updated, a null value indicates that nothing should be done.

        Args:
            time_series (list[stable.time_series.TimeSeries]):   List of time series data transfer objects to update.

        Returns:
            None

        Examples:
            Update the unit of a time series::

                from cognite.client.stable.time_series import TimeSeries
                client = CogniteClient()

                my_time_series = [TimeSeries(name="my_ts_1", unit="celsius")]

                client.time_series.update_time_series(my_time_series)
        """
        url = "/timeseries"
        items = [ts.camel_case_dict() for ts in time_series]
        body = {"items": items}
        self._put(url, body=body)

    def delete_time_series(self, name) -> None:
        """Delete a timeseries.

        Args:
            name (str):   Name of timeseries to delete.

        Returns:
            None

        Examples:
            Delete a time series by name::

                client = CogniteClient()

                client.time_series.delete_time_series(name="my_ts_1")
        """
        url = "/timeseries/{}".format(quote(name, safe=""))
        self._delete(url)



# -*- coding: utf-8 -*-
"""Utilites for Cognite API SDK

This module provides helper methods and different utilities for the Cognite API Python SDK.

This module is protected and should not used by end-users.
"""
import datetime
import platform
import re
import time
from datetime import datetime, timezone
from typing import Callable, List

import cognite.client


def datetime_to_ms(dt):
    return int(dt.replace(tzinfo=timezone.utc).timestamp() * 1000)


def granularity_to_ms(time_string):
    """Returns millisecond representation of granularity time string"""
    magnitude = int("".join([c for c in time_string if c.isdigit()]))
    unit = "".join([c for c in time_string if c.isalpha()])
    unit_in_ms = {
        "s": 1000,
        "second": 1000,
        "m": 60000,
        "minute": 60000,
        "h": 3600000,
        "hour": 3600000,
        "d": 86400000,
        "day": 86400000,
    }
    return magnitude * unit_in_ms[unit]


def _time_ago_to_ms(time_ago_string):
    """Returns millisecond representation of time-ago string"""
    if time_ago_string == "now":
        return 0
    pattern = r"(\d+)([a-z])-ago"
    res = re.match(pattern, str(time_ago_string))
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        unit_in_ms = {"s": 1000, "m": 60000, "h": 3600000, "d": 86400000, "w": 604800000}
        return magnitude * unit_in_ms[unit]
    return None


def interval_to_ms(start, end):
    """Returns the ms representation of start-end-interval whether it is time-ago, datetime or None."""
    time_now = int(round(time.time() * 1000))
    if isinstance(start, datetime):
        start = datetime_to_ms(start)
    elif isinstance(start, str):
        start = time_now - _time_ago_to_ms(start)
    elif start is None:
        start = time_now - _time_ago_to_ms("2w-ago")

    if isinstance(end, datetime):
        end = datetime_to_ms(end)
    elif isinstance(end, str):
        end = time_now - _time_ago_to_ms(end)
    elif end is None:
        end = time_now

    return start, end


class Bin:
    """
    Attributes:
        entries (List): List of entries.
        get_count (Callable): Callable function to get count.
    """

    def __init__(self, get_count):
        """
        Args:
            get_count: A function that will take an element and get the count of something in it.
        """
        self.entries = []
        self.get_count = get_count

    def add_item(self, item):
        self.entries.append(item)

    def sum(self):
        total = 0
        for elem in self.entries:
            total += self.get_count(elem)
        return total

    def show(self):
        return self.entries


def first_fit(list_items: List, max_size, get_count: Callable) -> List[List]:
    """Returns list of bins with input items inside."""

    # Sort the input list in decreasing order
    list_items = sorted(list_items, key=get_count, reverse=True)

    list_bins = [Bin(get_count=get_count)]

    for item in list_items:
        # Go through bins and try to allocate
        alloc_flag = False

        for bin in list_bins:
            if bin.sum() + get_count(item) <= max_size:
                bin.add_item(item)
                alloc_flag = True
                break

        # If item not allocated in bins in list, create new bin
        # and allocate it to it.
        if not alloc_flag:
            new_bin = Bin(get_count=get_count)
            new_bin.add_item(item)
            list_bins.append(new_bin)

    # Turn bins into list of items and return
    list_items = []
    for bin in list_bins:
        list_items.append(bin.show())

    return list_items


def get_user_agent():
    sdk_version = "CognitePythonSDK/{}".format(cognite.client.__version__)

    python_version = "{}/{} ({};{})".format(
        platform.python_implementation(), platform.python_version(), platform.python_build(), platform.python_compiler()
    )

    os_version_info = [platform.release(), platform.machine(), platform.architecture()[0]]
    os_version_info = [s for s in os_version_info if s]  # Ignore empty strings
    os_version_info = "-".join(os_version_info)
    operating_system = "{}/{}".format(platform.system(), os_version_info)

    return "{} {} {}".format(sdk_version, python_version, operating_system)


def _round_to_nearest(x, base):
    return int(base * round(float(x) / base))


def get_datapoints_windows(start: int, end: int, granularity: str, num_of_workers):
    diff = end - start
    granularity_ms = 1
    if granularity:
        granularity_ms = granularity_to_ms(granularity)

    # Ensure that number of steps is not greater than the number data points that will be returned
    steps = min(num_of_workers, max(1, int(diff / granularity_ms)))
    # Make step size a multiple of the granularity requested in order to ensure evenly spaced results
    step_size = _round_to_nearest(int(diff / steps), base=granularity_ms)
    # Create list of where each of the parallelized intervals will begin
    step_starts = [start + (i * step_size) for i in range(steps)]
    windows = [{"start": start, "end": start + step_size} for start in step_starts]
    if windows[-1]["end"] < end:
        windows[-1]["end"] = end
    return windows



# -*- coding: utf-8 -*-
import json
from copy import deepcopy
from typing import List

import pandas as pd

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResource, CogniteResponse


class EventResponse(CogniteResponse):
    """Event Response Object."""

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.id = item.get("id")
        self.type = item.get("type")
        self.sub_type = item.get("subType")
        self.description = item.get("description")
        self.asset_ids = item.get("assetIds")
        self.created_time = item.get("createdTime")
        self.start_time = item.get("startTime")
        self.end_time = item.get("endTime")
        self.last_updated_time = item.get("lastUpdatedTime")
        self.metadata = item.get("metadata")

    def to_pandas(self):
        event = self.to_json().copy()
        if event.get("metadata"):
            event.update(event.pop("metadata"))

        # Hack to avoid assetIds ending up as first element in dict as from_dict will fail
        asset_ids = event.pop("assetIds")
        df = pd.DataFrame.from_dict(event, orient="index")
        df.loc["assetIds"] = [asset_ids]
        return df


class EventListResponse(CogniteCollectionResponse):
    """Event List Response Object."""

    _RESPONSE_CLASS = EventResponse

    def __init__(self, internal_representation):
        super().__init__(internal_representation)

    def to_pandas(self):
        items = deepcopy(self.to_json())
        for d in items:
            if d.get("metadata"):
                d.update(d.pop("metadata"))
        return pd.DataFrame(items)


class Event(CogniteResource):
    """Data transfer object for events.

    Args:
        start_time (int):       Start time of the event in ms since epoch.
        end_time (int):         End time of the event in ms since epoch.
        description (str):      Textual description of the event.
        type (str):             Type of the event, e.g. 'failure'.
        subtype (str):          Subtype of the event, e.g. 'electrical'.
        metadata (dict):        Customizable extra data about the event.
        asset_ids (list[int]):  List of Asset IDs of related equipments that this event relates to.
    """

    def __init__(
        self, start_time=None, end_time=None, description=None, type=None, subtype=None, metadata=None, asset_ids=None
    ):
        self.start_time = start_time
        self.end_time = end_time
        self.description = description
        self.type = type
        self.subtype = subtype
        self.metadata = metadata
        self.asset_ids = asset_ids


class EventsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def get_event(self, event_id: int) -> EventResponse:
        """Returns a EventResponse containing an event matching the id.

        Args:
            event_id (int):         The event id.

        Returns:
            stable.events.EventResponse: A data object containing the requested event.

        Examples:
            Getting an event::

                client = CogniteClient()
                res = client.events.get_event(123)
                print(res)
        """
        url = "/events/{}".format(event_id)
        res = self._get(url)
        return EventResponse(res.json())

    def get_events(self, type=None, sub_type=None, asset_id=None, **kwargs) -> EventListResponse:
        """Returns an EventListReponse object containing events matching the query.

        Args:
            type (str):             Type (class) of event, e.g. 'failure'.
            sub_type (str):         Sub-type of event, e.g. 'electrical'.
            asset_id (int):         Return events associated with this assetId.
        Keyword Arguments:
            sort (str):             Sort descending or ascending. Default 'ASC'.
            cursor (str):           Cursor to use for paging through results.
            limit (int):            Return up to this many results. Maximum is 10000. Default is 25.
            has_description (bool): Return only events that have a textual description. Default null. False gives only
                                    those without description.
            min_start_time (string): Only return events from after this time.
            max_start_time (string): Only return events form before this time.
            autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be
                                    disregarded. Defaults to False.

        Returns:
            stable.events.EventListResponse: A data object containing the requested event.

        Examples:
            Getting all events of a given type::

                client = CogniteClient()
                res = client.events.get_events(type="a special type", autopaging=True)
                print(res.to_pandas())
        """
        autopaging = kwargs.get("autopaging", False)
        url = "/events"

        params = {
            "type": type,
            "subtype": sub_type,
            "assetId": asset_id,
            "sort": kwargs.get("sort"),
            "cursor": kwargs.get("cursor"),
            "limit": kwargs.get("limit", 25) if not autopaging else self._LIMIT,
            "hasDescription": kwargs.get("has_description"),
            "minStartTime": kwargs.get("min_start_time"),
            "maxStartTime": kwargs.get("max_start_time"),
        }

        res = self._get(url, params=params, autopaging=autopaging)
        return EventListResponse(res.json())

    def post_events(self, events: List[Event]) -> EventListResponse:
        """Adds a list of events and returns an EventListResponse object containing created events.

        Args:
            events (List[stable.events.Event]):    List of events to create.

        Returns:
            stable.events.EventListResponse

        Examples:
            Posting two events and linking them to an asset::

                from cognite.client.stable.events import Event
                client = CogniteClient()

                my_events = [Event(start_time=1, end_time=10, type="workorder", asset_ids=[123]),
                            Event(start_time=11, end_time=20, type="workorder", asset_ids=[123])]
                res = client.events.post_events(my_events)
                print(res)
        """
        url = "/events"
        items = [event.camel_case_dict() for event in events]
        body = {"items": items}
        res = self._post(url, body=body)
        return EventListResponse(res.json())

    def delete_events(self, event_ids: List[int]) -> None:
        """Deletes a list of events.

        Args:
            event_ids (List[int]):    List of ids of events to delete.

        Returns:
            None

        Examples:
            Deleting a list of events::

                client = CogniteClient()
                res = client.events.delete_events(event_ids=[1,2,3,4,5])
        """
        url = "/events/delete"
        body = {"items": event_ids}
        self._post(url, body=body)

    def search_for_events(
        self,
        description=None,
        type=None,
        subtype=None,
        min_start_time=None,
        max_start_time=None,
        min_end_time=None,
        max_end_time=None,
        min_created_time=None,
        max_created_time=None,
        min_last_updated_time=None,
        max_last_updated_time=None,
        metadata=None,
        asset_ids=None,
        asset_subtrees=None,
        **kwargs
    ):
        """Search for events.

        Args:
            description (str):   Prefix and fuzzy search on description.
            type (str):             Filter on type (case-sensitive).
            subtype (str):          Filter on subtype (case-sensitive).
            min_start_time (str):   Filter out events with startTime before this. Format is milliseconds since epoch.
            max_start_time (str):   Filter out events with startTime after this. Format is milliseconds since epoch.
            min_end_time (str):     Filter out events with endTime before this. Format is milliseconds since epoch.
            max_end_time (str):     Filter out events with endTime after this. Format is milliseconds since epoch.
            min_created_time(str):  Filter out events with createdTime before this. Format is milliseconds since epoch.
            max_created_time (str): Filter out events with createdTime after this. Format is milliseconds since epoch.
            min_last_updated_time(str):  Filter out events with lastUpdatedtime before this. Format is milliseconds since epoch.
            max_last_updated_time(str): Filter out events with lastUpdatedtime after this. Format is milliseconds since epoch.
            metadata (dict):        Filter out events that do not match these metadata fields and values (case-sensitive).
                                    Format is {"key1":"value1","key2":"value2"}.
            asset_ids (List[int]):  Filter out events that are not linked to any of these assets. Format is [12,345,6,7890].
            asset_subtrees (List[int]): Filter out events that are not linked to assets in the subtree rooted at these assets.
                                        Format is [12,345,6,7890].

        Keyword Args:
            sort (str):             Field to be sorted.
            dir (str):              Sort direction (desc or asc)
            limit (int):            Return up to this many results. Max is 1000, default is 25.
            offset (int):           Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.

        Returns:
            stable.events.EventListResponse.

        Examples:
            Perform a fuzzy search on event descriptions and get the first 3 results::

                client = CogniteClient()
                res = client.events.search_for_events(description="Something like this", limit=10)
                print(res)
        """
        url = "/events/search"
        params = {
            "description": description,
            "type": type,
            "subtype": subtype,
            "minStartTime": min_start_time,
            "maxStartTime": max_start_time,
            "minEndTime": min_end_time,
            "maxEndTime": max_end_time,
            "minCreatedTime": min_created_time,
            "maxCreatedTime": max_created_time,
            "minLastUpdatedTime": min_last_updated_time,
            "maxLastUpdatedTime": max_last_updated_time,
            "metadata": json.dumps(metadata),
            "assetIds": str(asset_ids or []),
            "assetSubtrees": str(asset_subtrees) if asset_subtrees else None,
            "sort": kwargs.get("sort"),
            "dir": kwargs.get("dir"),
            "limit": kwargs.get("limit", self._LIMIT),
            "offset": kwargs.get("offset"),
        }

        res = self._get(url, params=params)
        return EventListResponse(res.json())



import functools
import gzip
import json
import logging
import os
import re
from typing import Any, Dict

import numpy
from requests import Response, Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3 import Retry

from cognite.client.exceptions import APIError

log = logging.getLogger("cognite-sdk")

DEFAULT_NUM_OF_RETRIES = 5
HTTP_METHODS_TO_RETRY = [429, 500, 502, 503]


def _init_requests_session():
    session = Session()
    num_of_retries = int(os.getenv("COGNITE_NUM_RETRIES", DEFAULT_NUM_OF_RETRIES))
    retry = Retry(
        total=num_of_retries,
        read=num_of_retries,
        connect=num_of_retries,
        backoff_factor=0.5,
        status_forcelist=HTTP_METHODS_TO_RETRY,
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


_REQUESTS_SESSION = _init_requests_session()


def _status_is_valid(status_code: int):
    return status_code < 400


def _raise_API_error(res: Response):
    x_request_id = res.headers.get("X-Request-Id")
    code = res.status_code
    extra = {}
    try:
        error = res.json()["error"]
        if isinstance(error, str):
            msg = error
        else:
            msg = error["message"]
            extra = error.get("extra")
    except:
        msg = res.content

    log.error("HTTP Error %s: %s", code, msg, extra={"X-Request-ID": x_request_id, "extra": extra})
    raise APIError(msg, code, x_request_id, extra=extra)


def _log_request(res: Response, **kwargs):
    method = res.request.method
    url = res.request.url
    status_code = res.status_code

    extra = kwargs.copy()
    extra["headers"] = res.request.headers
    if "api-key" in extra.get("headers", {}):
        extra["headers"]["api-key"] = None

    http_protocol_version = ".".join(list(str(res.raw.version)))

    log.info("HTTP/{} {} {} {}".format(http_protocol_version, method, url, status_code), extra=extra)


def request_method(method=None):
    @functools.wraps(method)
    def wrapper(client_instance, url, *args, **kwargs):
        if not url.startswith("/"):
            raise ValueError("URL must start with '/'")
        full_url = client_instance._base_url + url

        # Hack to allow running model hosting requests against local emulator
        if os.getenv("USE_MODEL_HOSTING_EMULATOR") == "1":
            full_url = _model_hosting_emulator_url_converter(full_url)

        default_headers = client_instance._headers.copy()
        default_headers.update(kwargs.get("headers") or {})
        kwargs["headers"] = default_headers
        res = method(client_instance, full_url, *args, **kwargs)
        if _status_is_valid(res.status_code):
            return res
        _raise_API_error(res)

    return wrapper


def _model_hosting_emulator_url_converter(url):
    pattern = "https://api.cognitedata.com/api/0.6/projects/(.*)/analytics/models(.*)"
    res = re.match(pattern, url)
    if res is not None:
        project = res.group(1)
        path = res.group(2)
        return "http://localhost:8000/api/0.1/projects/{}/models{}".format(project, path)
    return url


class APIClient:
    _LIMIT = 1000

    def __init__(
        self,
        version: str = None,
        project: str = None,
        base_url: str = None,
        num_of_workers: int = None,
        cookies: Dict = None,
        headers: Dict = None,
        timeout: int = None,
    ):
        self._request_session = _REQUESTS_SESSION
        self._project = project
        __base_path = "/api/{}/projects/{}".format(version, project) if version else ""
        self._base_url = base_url + __base_path
        self._num_of_workers = num_of_workers
        self._cookies = cookies
        self._headers = headers
        self._timeout = timeout

    @request_method
    def _delete(self, url: str, params: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        res = self._request_session.delete(
            url, params=params, headers=headers, cookies=self._cookies, timeout=self._timeout
        )
        _log_request(res)
        return res

    def _autopaged_get(self, url: str, params: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        params = params.copy()
        items = []
        while True:
            url = re.sub("{}(/.*)".format(self._base_url), r"\1", url)
            res = self._get(url, params=params, headers=headers)
            params["cursor"] = res.json()["data"].get("nextCursor")
            items.extend(res.json()["data"]["items"])
            next_cursor = res.json()["data"].get("nextCursor")
            if not next_cursor:
                break
        res._content = json.dumps({"data": {"items": items}}).encode()
        return res

    @request_method
    def _get(self, url: str, params: Dict[str, Any] = None, headers: Dict[str, Any] = None, autopaging: bool = False):
        if autopaging:
            return self._autopaged_get(url, params, headers)
        res = self._request_session.get(
            url, params=params, headers=headers, cookies=self._cookies, timeout=self._timeout
        )
        _log_request(res)
        return res

    @request_method
    def _post(self, url: str, body: Dict[str, Any], params: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        data = json.dumps(body, default=self._json_dumps_default)
        if not os.getenv("COGNITE_DISABLE_GZIP", False):
            headers["Content-Encoding"] = "gzip"
            data = gzip.compress(data.encode())
        res = self._request_session.post(
            url, data=data, headers=headers, params=params, cookies=self._cookies, timeout=self._timeout
        )
        _log_request(res, body=body)
        return res

    @request_method
    def _put(self, url: str, body: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        data = json.dumps(body or {}, default=self._json_dumps_default)
        if not os.getenv("COGNITE_DISABLE_GZIP", False):
            headers["Content-Encoding"] = "gzip"
            data = gzip.compress(data.encode())
        res = self._request_session.put(url, data=data, headers=headers, cookies=self._cookies, timeout=self._timeout)
        _log_request(res, body=body)
        return res

    @staticmethod
    def _json_dumps_default(x):
        if isinstance(x, numpy.int_):
            return int(x)
        if isinstance(x, numpy.float_):
            return float(x)
        if isinstance(x, numpy.bool_):
            return bool(x)
        return x.__dict__


class CogniteResponse:
    """Cognite Response class

    All responses inherit from this class.

    Examples:
        All responses are pretty-printable::

            from cognite.client import CogniteClient

            client = CogniteClient()
            res = client.assets.get_assets(limit=1)

            print(res)

        All endpoints which support paging have an ``autopaging`` flag which may be set to true in order to sequentially
        fetch all resources. If for some reason, you want to do this manually, you may use the next_cursor() method on
        the response object. Here is an example of that::

            from cognite.client import CogniteClient

            client = CogniteClient()

            asset_list = []

            cursor = None
            while True:
                res = client.assets.get_assets(cursor=cursor)
                asset_list.extend(res.to_json())
                cursor = res.next_cursor()
                if cursor is None:
                    break

            print(asset_list)
    """

    def __init__(self, internal_representation):
        self.internal_representation = internal_representation

    def __str__(self):
        return json.dumps(self.to_json(), indent=4, sort_keys=True)

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"][0]

    def next_cursor(self):
        """Returns next cursor to use for paging through results. Returns ``None`` if there are no more results."""
        if self.internal_representation.get("data"):
            return self.internal_representation.get("data").get("nextCursor")

    def previous_cursor(self):
        """Returns previous cursor to use for paging through results. Returns ``None`` if there are no more results."""
        if self.internal_representation.get("data"):
            return self.internal_representation.get("data").get("previousCursor")


class CogniteCollectionResponse(CogniteResponse):
    """Cognite Collection Response class

    All collection responses inherit from this class. Collection responses are subscriptable and iterable.
    """

    _RESPONSE_CLASS = None

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"]

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__({"data": {"items": self.to_json()[index]}})
        return self._RESPONSE_CLASS({"data": {"items": [self.to_json()[index]]}})

    def __len__(self):
        return len(self.to_json())

    def __iter__(self):
        self.counter = 0
        return self

    def __next__(self):
        if self.counter > len(self.to_json()) - 1:
            raise StopIteration
        else:
            self.counter += 1
            return self._RESPONSE_CLASS({"data": {"items": [self.to_json()[self.counter - 1]]}})


class CogniteResource:
    @staticmethod
    def _to_camel_case(snake_case_string: str):
        components = snake_case_string.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    def camel_case_dict(self):
        new_d = {}
        for key in self.__dict__:
            new_d[self._to_camel_case(key)] = self.__dict__[key]
        return new_d

    def to_json(self):
        return json.loads(json.dumps(self, default=lambda x: x.__dict__))

    def __eq__(self, other):
        return type(self) == type(other) and self.to_json() == other.to_json()

    def __str__(self):
        return json.dumps(self.__dict__, default=lambda x: x.__dict__, indent=4, sort_keys=True)



# -*- coding: utf-8 -*-
import json
from typing import List

import pandas as pd

from cognite.client._api_client import APIClient, CogniteResponse


class RawResponse(CogniteResponse):
    """Raw Response Object."""

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        return pd.DataFrame(self.to_json())


class RawRow(object):
    """DTO for a row in a raw database.

    The Raw API is a simple key/value-store. Each row in a table in a raw database consists of a
    unique row key and a set of columns.

    Args:
        key (str):      Unique key for the row.

        columns (dict):  A key/value-map consisting of the values in the row.
    """

    def __init__(self, key, columns):
        self.key = key
        self.columns = columns

    def __repr__(self):
        return json.dumps(self.repr_json())

    def repr_json(self):
        return self.__dict__


class RawClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def get_databases(self, limit: int = None, cursor: str = None) -> RawResponse:
        """Returns a RawObject containing a list of raw databases.

        Args:
            limit (int):    A limit on the amount of results to return.

            cursor (str):   A cursor can be provided to navigate through pages of results.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/raw"
        params = {"limit": limit, "cursor": cursor}
        res = self._get(url=url, params=params, headers={"content-type": "*/*"})
        return RawResponse(res.json())

    def create_databases(self, database_names: list) -> RawResponse:
        """Creates databases in the Raw API and returns the created databases.

        Args:
            database_names (list):  A list of databases to create.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.

        """
        url = "/raw/create"
        body = {"items": [{"dbName": "{}".format(database_name)} for database_name in database_names]}
        res = self._post(url=url, body=body, headers={"content-type": "*/*"})
        return RawResponse(res.json())

    def delete_databases(self, database_names: list, recursive: bool = False) -> None:
        """Deletes databases in the Raw API.

        Args:
            database_names (list):  A list of databases to delete.

        Returns:
            None

        """
        url = "/raw/delete"
        body = {"items": [{"dbName": "{}".format(database_name)} for database_name in database_names]}
        params = {"recursive": recursive}
        self._post(url=url, body=body, params=params, headers={"content-type": "*/*"})

    def get_tables(self, database_name: str = None, limit: int = None, cursor: str = None) -> RawResponse:
        """Returns a RawObject containing a list of tables in a raw database.

        Args:
            database_name (str):   The database name to retrieve tables from.

            limit (int):    A limit on the amount of results to return.

            cursor (str):   A cursor can be provided to navigate through pages of results.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/raw/{}".format(database_name)
        params = dict()
        if limit is not None:
            params["limit"] = limit
        if cursor is not None:
            params["cursor"] = cursor
        res = self._get(url=url, params=params, headers={"content-type": "*/*"})
        return RawResponse(res.json())

    def create_tables(self, database_name: str = None, table_names: list = None) -> RawResponse:
        """Creates tables in the given Raw API database.

        Args:
            database_name (str):    The database to create tables in.

            table_names (list):     The table names to create.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.

        """
        url = "/raw/{}/create".format(database_name)
        body = {"items": [{"tableName": "{}".format(table_name)} for table_name in table_names]}
        res = self._post(url=url, body=body, headers={"content-type": "*/*"})
        return RawResponse(res.json())

    def delete_tables(self, database_name: str = None, table_names: list = None) -> None:
        """Deletes databases in the Raw API.

        Args:
            database_name (str):    The database to create tables in.

            table_names (list):     The table names to create.

        Returns:
            None

        """
        url = "/raw/{}/delete".format(database_name)
        body = {"items": [{"tableName": "{}".format(table_name)} for table_name in table_names]}
        self._post(url=url, body=body, headers={"content-type": "*/*"})

    def get_rows(
        self, database_name: str = None, table_name: str = None, limit: int = None, cursor: str = None
    ) -> RawResponse:
        """Returns a RawObject containing a list of rows.

        Args:
            database_name (str):    The database name to retrieve rows from.

            table_name (str):       The table name to retrieve rows from.

            limit (int):            A limit on the amount of results to return.

            cursor (str):           A cursor can be provided to navigate through pages of results.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/raw/{}/{}".format(database_name, table_name)
        params = dict()
        params["limit"] = limit
        params["cursor"] = cursor
        res = self._get(url=url, params=params, headers={"content-type": "*/*"})
        return RawResponse(res.json())

    def create_rows(
        self, database_name: str = None, table_name: str = None, rows: List[RawRow] = None, ensure_parent=False
    ) -> None:
        """Creates tables in the given Raw API database.

        Args:
            database_name (str):    The database to create rows in.

            table_name (str):       The table names to create rows in.

            rows (list[stable.raw.RawRow]):            The rows to create.

            ensure_parent (bool):   Create database/table if it doesn't exist already

        Returns:
            None

        """
        url = "/raw/{}/{}/create".format(database_name, table_name)

        params = {}
        if ensure_parent:
            params = {"ensureParent": "true"}

        ul_row_limit = 1000
        i = 0
        while i < len(rows):
            body = {
                "items": [{"key": "{}".format(row.key), "columns": row.columns} for row in rows[i : i + ul_row_limit]]
            }
            self._post(url=url, body=body, headers={"content-type": "*/*"}, params=params)
            i += ul_row_limit

    def delete_rows(self, database_name: str = None, table_name: str = None, rows: List[RawRow] = None) -> None:
        """Deletes rows in the Raw API.

        Args:
            database_name (str):    The database to create tables in.

            table_name (str):      The table name where the rows are at.

            rows (list):            The rows to delete.

        Returns:
            None

        """
        url = "/raw/{}/{}/delete".format(database_name, table_name)
        body = {"items": [{"key": "{}".format(row.key), "columns": row.columns} for row in rows]}
        self._post(url=url, body=body, headers={"content-type": "*/*"})

    def get_row(self, database_name: str = None, table_name: str = None, row_key: str = None) -> RawResponse:
        """Returns a RawObject containing a list of rows.

        Args:
            database_name (str):    The database name to retrieve rows from.

            table_name (str):       The table name to retrieve rows from.

            row_key (str):          The key of the row to fetch.

        Returns:
            stable.raw.RawResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/raw/{}/{}/{}".format(database_name, table_name, row_key)
        params = dict()
        res = self._get(url=url, params=params, headers={"content-type": "*/*"})
        return RawResponse(res.json())



from cognite.client.experimental.model_hosting import ModelHostingClient
from cognite.client.experimental.datapoints import DatapointsClient
from cognite.client.experimental.sequences import SequencesClient
from cognite.client.experimental.time_series import TimeSeriesClient


class ExperimentalClient:
    def __init__(self, client_factory):
        self._client_factory = client_factory
        self._model_hosting_client = ModelHostingClient(client_factory=self._client_factory)
        self._datapoints_client = self._client_factory(DatapointsClient)
        self._sequences_client = self._client_factory(SequencesClient)
        self._time_series_client = self._client_factory(TimeSeriesClient)

    @property
    def model_hosting(self) -> ModelHostingClient:
        return self._model_hosting_client

    @property
    def datapoints(self) -> DatapointsClient:
        return self._datapoints_client

    @property
    def sequences(self) -> SequencesClient:
        return self._sequences_client

    @property
    def time_series(self) -> TimeSeriesClient:
        return self._time_series_client



# -*- coding: utf-8 -*-
import json
from typing import Dict, List

import pandas as pd

from cognite.client._api_client import APIClient, CogniteCollectionResponse, CogniteResource, CogniteResponse


class AssetResponse(CogniteResponse):
    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.id = item.get("id")
        self.name = item.get("name")
        self.depth = item.get("depth")
        self.description = item.get("description")
        self.created_time = item.get("createdTime")
        self.last_updated_time = item.get("lastUpdatedTime")
        self.metadata = item.get("metadata")
        self.parent_id = item.get("parentId")
        self.path = item.get("path")

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        if len(self.to_json()) > 0:
            asset = self.to_json().copy()
            # Hack to avoid path ending up as first element in dict as from_dict will fail
            path = asset.pop("path")
            df = pd.DataFrame.from_dict(asset, orient="index")
            df.loc["path"] = [path]
            return df
        return pd.DataFrame()


class AssetListResponse(CogniteCollectionResponse):
    """Assets Response Object"""

    _RESPONSE_CLASS = AssetResponse

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        if len(self.to_json()) > 0:
            return pd.DataFrame(self.internal_representation["data"]["items"])
        return pd.DataFrame()


class Asset(CogniteResource):
    """Data transfer object for assets.

    Args:
        name (str):                 Name of asset. Often referred to as tag.
        id (int):                   Id of asset.
        parent_id (int):            ID of parent asset, if any.
        description (str):          Description of asset.
        metadata (dict):            Custom , application specific metadata. String key -> String Value.
        ref_id (str):               Reference ID used only in post request to disambiguate references to duplicate
                                    names.
        parent_name (str):          Name of parent, this parent must exist in the same POST request.
        parent_ref_id (str):        Reference ID of parent, to disambiguate if multiple nodes have the same name.
        source (str):               Source of asset.
        source_id (str):            Source id of asset.
    """

    def __init__(
        self,
        name=None,
        id=None,
        parent_id=None,
        description=None,
        metadata=None,
        ref_id=None,
        parent_name=None,
        parent_ref_id=None,
        source=None,
        source_id=None,
    ):
        self.name = name
        self.id = id
        self.parent_id = parent_id
        self.description = description
        self.metadata = metadata
        self.ref_id = ref_id
        self.parent_name = parent_name
        self.parent_ref_id = parent_ref_id
        self.source = source
        self.source_id = source_id


class AssetsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def get_assets(
        self, name=None, path=None, description=None, metadata=None, depth=None, fuzziness=None, **kwargs
    ) -> AssetListResponse:
        """Returns assets matching provided description.

        Args:
            name (str):             The name of the asset(s) to get.

            path (List[int]):       The path of the subtree to search in.

            description (str):      Search query.

            metadata (dict):         The metadata values used to filter the results.

            depth (int):            Get sub assets up oto this many levels below the specified path.

            fuzziness (int):        The degree of fuzziness in the name matching.

        Keyword Arguments:
            autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be
                                    disregarded. Defaults to False.

            limit (int):            The maximum number of assets to be returned.

            cursor (str):           Cursor to use for paging through results.
        Returns:
            stable.assets.AssetListResponse: A data object containing the requested assets with several getter methods with different
            output formats.

        Examples:
            You can fetch all assets with a maximum depth of three like this::

                client = CogniteClient()
                res = client.assets.get_assets(depth=3, autopaging=True)
                print(res.to_pandas())

            You can fetch all assets in a given subtree like this::

                client = CogniteClient()
                res = client.assets.get_assets(path=[1,2,3], autopaging=True)
                print(res.to_pandas())
        """
        autopaging = kwargs.get("autopaging", False)
        url = "/assets"
        params = {
            "name": name,
            "description": description,
            "path": str(path) if path else None,
            "metadata": str(metadata) if metadata else None,
            "depth": depth,
            "fuzziness": fuzziness,
            "cursor": kwargs.get("cursor"),
            "limit": kwargs.get("limit", self._LIMIT) if not autopaging else self._LIMIT,
        }
        res = self._get(url, params=params, autopaging=autopaging)
        return AssetListResponse(res.json())

    def get_asset(self, asset_id) -> AssetResponse:
        """Returns the asset with the provided assetId.

        Args:
            asset_id (int):         The asset id of the top asset to get.

        Returns:
            stable.assets.AssetResponse: A data object containing the requested assets with several getter methods with different
            output formats.
        Examples:
            You can fetch a single asset like this::

                client = CogniteClient()
                res = client.assets.get_asset(asset_id=123)
                print(res)
        """
        url = "/assets/{}/subtree".format(asset_id)
        res = self._get(url)
        return AssetResponse(res.json())

    def get_asset_subtree(self, asset_id, depth=None, **kwargs) -> AssetListResponse:
        """Returns asset subtree of asset with provided assetId.

        Args:
            asset_id (int):         The asset id of the top asset to get.

            depth (int):            Get subassets this many levels below the top asset.

        Keyword Arguments:
            limit (int):            The maximum nuber of assets to be returned.

            cursor (str):           Cursor to use for paging through results.

            autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be
                                    disregarded. Defaults to False.
        Returns:
            stable.assets.AssetListResponse: A data object containing the requested assets with several getter methods
            with different output formats.

        Examples:
            You can fetch an asset subtree like this::

                client = CogniteClient()
                res = client.assets.get_asset_subtree(asset_id=123, depth=2)
                print(res.to_pandas())
        """
        autopaging = kwargs.get("autopaging", False)
        url = "/assets/{}/subtree".format(asset_id)
        params = {
            "depth": depth,
            "limit": kwargs.get("limit", self._LIMIT) if not autopaging else self._LIMIT,
            "cursor": kwargs.get("cursor"),
        }
        res = self._get(url, params=params, autopaging=autopaging)
        return AssetListResponse(res.json())

    def post_assets(self, assets: List[Asset]) -> AssetListResponse:
        """Insert a list of assets.

        Args:
            assets (list[stable.assets.Asset]): List of asset data transfer objects.

        Returns:
            stable.assets.AssetListResponse: A data object containing the posted assets with several getter methods with different
            output formats.

        Examples:
            Posting an asset::

                from cognite.client.stable.assets import Asset

                client = CogniteClient()

                my_asset = Asset("myasset")
                assets_to_post = [my_asset]
                res = client.assets.post_assets(assets_to_post)
                print(res)
        """
        url = "/assets"
        items = [asset.camel_case_dict() for asset in assets]
        res = self._post(url, body={"items": items})
        return AssetListResponse(res.json())

    def delete_assets(self, asset_ids: List[int]) -> None:
        """Delete a list of assets.

        Args:
            asset_ids (list[int]): List of IDs of assets to delete.

        Returns:
            None

        Examples:
            You can delete an asset like this::

                client = CogniteClient()
                res = client.assets.delete_assets([123])
        """
        url = "/assets/delete"
        body = {"items": asset_ids}
        self._post(url, body=body)

    @staticmethod
    def _asset_to_patch_format(id, name=None, description=None, metadata=None, source=None, source_id=None):
        patch_asset = {"id": id}
        if name:
            patch_asset["name"] = {"set": name}
        if description:
            patch_asset["description"] = {"set": description}
        if metadata:
            patch_asset["metadata"] = {"set": metadata}
        if source:
            patch_asset["source"] = {"set": source}
        if source_id:
            patch_asset["sourceId"] = {"set": source_id}
        return patch_asset

    def update_asset(
        self,
        asset_id: int,
        name: str = None,
        description: str = None,
        metadata: Dict = None,
        source: str = None,
        source_id: str = None,
    ) -> AssetResponse:
        """Update an asset

        Args:
            asset_id (int): The id of the asset to update
            name (str, optional): The new name
            description (str, optional): The new description
            metadata (Dict, optional): The new metadata
            source (str, optional): The new source
            source_id (str, optional): The new source id

        Returns:
            AssetResponse: The updated asset
        """
        url = "/assets/{}/update".format(asset_id)
        body = self._asset_to_patch_format(asset_id, name, description, metadata, source, source_id)
        res = self._post(url, body=body)
        return AssetResponse(res.json())

    def update_assets(self, assets: List[Asset]):
        """Update multiple assets

        Args:
            assets (List[stable.assets.Asset]): List of assets to update

        Returns:
            AssetListResponse: List of updated assets
        """
        url = "/assets/update"
        items = [
            self._asset_to_patch_format(a.id, a.name, a.description, a.metadata, a.source, a.source_id) for a in assets
        ]
        res = self._post(url, body={"items": items})
        return AssetListResponse(res.json())

    def search_for_assets(
        self,
        name=None,
        description=None,
        query=None,
        metadata=None,
        asset_subtrees=None,
        min_created_time=None,
        max_created_time=None,
        min_last_updated_time=None,
        max_last_updated_time=None,
        **kwargs
    ) -> AssetListResponse:
        """Search for assets.

        Args:
            name:   Prefix and fuzzy search on name.
            description str:   Prefix and fuzzy search on description.
            query (str):       Search on name and description using wildcard search on each of the words
                                (separated by spaces). Retrieves results where at least one word must match.
                                Example: 'some other'
            metadata (dict):        Filter out assets that do not match these metadata fields and values (case-sensitive).
                                    Format is {"key1":"value1","key2":"value2"}.
            asset_subtrees (List[int]): Filter out assets that are not linked to assets in the subtree rooted at these assets.
                                        Format is [12,345,6,7890].
            min_created_time(str):  Filter out assets with createdTime before this. Format is milliseconds since epoch.
            max_created_time (str): Filter out assets with createdTime after this. Format is milliseconds since epoch.
            min_last_updated_time(str):  Filter out assets with lastUpdatedtime before this. Format is milliseconds since epoch.
            max_last_updated_time(str): Filter out assets with lastUpdatedtime after this. Format is milliseconds since epoch.

        Keyword Args:
            sort (str):             Field to be sorted.
            dir (str):              Sort direction (desc or asc)
            limit (int):            Return up to this many results. Max is 1000, default is 25.
            offset (int):           Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.
            boost_name (str):       Whether or not boosting name field. This option is test_experimental and can be changed.
        Returns:
            stable.assets.AssetListResponse.

        Examples:
            Searching for assets::

                client = CogniteClient()
                res = client.assets.search_for_assets(name="myasset")
                print(res)
        """
        url = "/assets/search"
        params = {
            "name": name,
            "description": description,
            "query": query,
            "metadata": json.dumps(metadata),
            "assetSubtrees": str(asset_subtrees) if asset_subtrees is not None else None,
            "minCreatedTime": min_created_time,
            "maxCreatedTime": max_created_time,
            "minLastUpdatedTime": min_last_updated_time,
            "maxLastUpdatedTime": max_last_updated_time,
            "sort": kwargs.get("sort"),
            "dir": kwargs.get("dir"),
            "limit": kwargs.get("limit", self._LIMIT),
            "offset": kwargs.get("offset"),
            "boostName": kwargs.get("boost_name"),
        }

        res = self._get(url, params=params)
        return AssetListResponse(res.json())



import os
import sys
from typing import Any, Dict

import requests

from cognite.client._api_client import APIClient
from cognite.client._utils import get_user_agent
from cognite.client.experimental import ExperimentalClient
from cognite.client.stable.assets import AssetsClient
from cognite.client.stable.datapoints import DatapointsClient
from cognite.client.stable.events import EventsClient
from cognite.client.stable.files import FilesClient
from cognite.client.stable.login import LoginClient
from cognite.client.stable.raw import RawClient
from cognite.client.stable.tagmatching import TagMatchingClient
from cognite.client.stable.time_series import TimeSeriesClient
from cognite.logger import configure_logger

DEFAULT_BASE_URL = "https://api.cognitedata.com"
DEFAULT_NUM_OF_WORKERS = 1
DEFAULT_TIMEOUT = 30


class CogniteClient:
    """Main entrypoint into Cognite Python SDK.

    All services are made available through this object. See examples below.

    Args:
        api_key (str): API key
        project (str): Project. Defaults to project of given API key.
        base_url (str): Base url to send requests to. Defaults to "https://api.cognitedata.com"
        num_of_workers (int): Number of workers to spawn when parallelizing data fetching. Defaults to 10.
        cookies (Dict): Cookies to append to all requests. Defaults to {}
        headers (Dict): Additional headers to add to all requests. Defaults are:
                 {"api-key": self.api_key, "content-type": "application/json", "accept": "application/json"}
        timeout (int): Timeout on requests sent to the api. Defaults to 60 seconds.
        debug (bool): Configures logger to log extra request details to stdout.


    Examples:
            The CogniteClient is instantiated and used like this. This example assumes that the environment variable
            COGNITE_API_KEY has been set::

                from cognite.client import CogniteClient
                client = CogniteClient()
                res = client.time_series.get_time_series()
                print(res.to_pandas())

            Certain experimental features are made available through this client as follows::

                from cognite.client import CogniteClient
                client = CogniteClient()
                res = client.experimental.model_hosting.models.list_models()
                print(res)

            Default configurations may be set using the following environment variables::

                export COGNITE_API_KEY = <your-api-key>
                export COGNITE_BASE_URL = http://<host>:<port>
                export COGNITE_NUM_RETRIES = <number-of-retries>
                export COGNITE_NUM_WORKERS = <number-of-workers>
                export COGNITE_TIMEOUT = <num-of-seconds>
                export COGNITE_DISABLE_GZIP = "1"
    """

    def __init__(
        self,
        api_key: str = None,
        project: str = None,
        base_url: str = None,
        num_of_workers: int = None,
        headers: Dict[str, str] = None,
        cookies: Dict[str, str] = None,
        timeout: int = None,
        debug: bool = None,
    ):
        thread_local_api_key, thread_local_project = self._get_thread_local_credentials()

        environment_api_key = os.getenv("COGNITE_API_KEY")
        environment_base_url = os.getenv("COGNITE_BASE_URL")
        environment_num_of_workers = os.getenv("COGNITE_NUM_WORKERS")
        environment_timeout = os.getenv("COGNITE_TIMEOUT")

        self.__api_key = api_key or thread_local_api_key or environment_api_key
        if self.__api_key is None:
            raise ValueError("No Api Key has been specified")

        self._base_url = base_url or environment_base_url or DEFAULT_BASE_URL

        self._num_of_workers = int(num_of_workers or environment_num_of_workers or DEFAULT_NUM_OF_WORKERS)

        self._configure_headers(headers)

        self._cookies = cookies or {}

        self._timeout = int(timeout or environment_timeout or DEFAULT_TIMEOUT)

        self._project = project or thread_local_project
        self._login_client = self._client_factory(LoginClient)
        if self._project is None:
            self._project = self._login_client.status().project

        self._api_client = self._client_factory(APIClient)

        if debug:
            configure_logger("cognite-sdk", log_level="INFO", log_json=True)

        self._assets_client = self._client_factory(AssetsClient)
        self._datapoints_client = self._client_factory(DatapointsClient)
        self._events_client = self._client_factory(EventsClient)
        self._files_client = self._client_factory(FilesClient)
        self._raw_client = self._client_factory(RawClient)
        self._tagmatching_client = self._client_factory(TagMatchingClient)
        self._time_series_client = self._client_factory(TimeSeriesClient)
        self._experimental_client = ExperimentalClient(self._client_factory)

    @property
    def assets(self) -> AssetsClient:
        return self._assets_client

    @property
    def datapoints(self) -> DatapointsClient:
        return self._datapoints_client

    @property
    def events(self) -> EventsClient:
        return self._events_client

    @property
    def files(self) -> FilesClient:
        return self._files_client

    @property
    def login(self) -> LoginClient:
        return self._login_client

    @property
    def raw(self) -> RawClient:
        return self._raw_client

    @property
    def tag_matching(self) -> TagMatchingClient:
        return self._tagmatching_client

    @property
    def time_series(self) -> TimeSeriesClient:
        return self._time_series_client

    @property
    def experimental(self) -> ExperimentalClient:
        return self._experimental_client

    def get(self, url: str, params: Dict[str, Any] = None, headers: Dict[str, Any] = None, autopaging: bool = False):
        """Perform a GET request to a path in the API.

        Comes in handy if the endpoint you want to reach is not currently supported by the SDK.
        """
        return self._api_client._get(url, params=params, headers=headers, autopaging=autopaging)

    def post(self, url: str, body: Dict[str, Any], params: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        """Perform a POST request to a path in the API.

        Comes in handy if the endpoint you want to reach is not currently supported by the SDK.
        """
        return self._api_client._post(url, body=body, params=params, headers=headers)

    def put(self, url: str, body: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        """Perform a PUT request to a path in the API.

        Comes in handy if the endpoint you want to reach is not currently supported by the SDK.
        """
        return self._api_client._put(url, body=body, headers=headers)

    def delete(self, url: str, params: Dict[str, Any] = None, headers: Dict[str, Any] = None):
        """Perform a DELETE request to a path in the API.

        Comes in handy if the endpoint you want to reach is not currently supported by the SDK.
        """
        return self._api_client._delete(url, params=params, headers=headers)

    def _client_factory(self, client):
        return client(
            project=self._project,
            base_url=self._base_url,
            num_of_workers=self._num_of_workers,
            cookies=self._cookies,
            headers=self._headers,
            timeout=self._timeout,
        )

    def _configure_headers(self, user_defined_headers):
        self._headers = requests.utils.default_headers()
        self._headers.update(
            {"api-key": self.__api_key, "content-type": "application/json", "accept": "application/json"}
        )

        if "User-Agent" in self._headers:
            self._headers["User-Agent"] += " " + get_user_agent()
        else:
            self._headers["User-Agent"] = get_user_agent()

        if user_defined_headers:
            self._headers.update(user_defined_headers)

    def _get_thread_local_credentials(self):
        if "cognite._thread_local" in sys.modules:
            from cognite._thread_local import credentials

            thread_local_api_key = getattr(credentials, "api_key", None)
            thread_local_project = getattr(credentials, "project", None)
            return thread_local_api_key, thread_local_project
        return None, None



# -*- coding: utf-8 -*-
import io
import json
import time
from concurrent.futures import ThreadPoolExecutor as Pool
from copy import copy
from datetime import datetime
from functools import partial
from typing import Iterable, List
from urllib.parse import quote

import pandas as pd

from cognite.client import _utils
from cognite.client._api_client import APIClient, CogniteResource, CogniteResponse
from cognite.client._auxiliary._protobuf_descriptors import _api_timeseries_data_v2_pb2


class DatapointsResponse(CogniteResponse):
    """Datapoints Response Object."""

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        item = self.to_json()
        self.name = item.get("name")
        self.datapoints = item.get("datapoints")

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"][0]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        return pd.DataFrame(self.internal_representation["data"]["items"][0]["datapoints"])


class DatapointsQuery(CogniteResource):
    """Data Query Object for Datapoints.

    Args:
        name (str):           Unique name of the time series.
        aggregates (list):          The aggregate functions to be returned. Use default if null. An empty string must
                                    be sent to get raw data if the default is a set of aggregate functions.
        granularity (str):          The granularity size and granularity of the aggregates.
        start (str, int, datetime): Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    Example: '2d-ago' will get everything that is up to 2 days old. Can also send time in
                                    ms since epoch or as a datetime object.
        end (str, int, datetime):   Get datapoints up to this time. The format is the same as for start.
    """

    def __init__(self, name, aggregates=None, granularity=None, start=None, end=None, limit=None):
        self.name = name
        self.aggregates = ",".join(aggregates) if aggregates is not None else None
        self.granularity = granularity
        self.start, self.end = _utils.interval_to_ms(start, end)
        if not start:
            self.start = None
        if not end:
            self.end = None
        self.limit = limit

    def __str__(self):
        return json.dumps(self.__dict__, indent=4, sort_keys=True)


class DatapointsResponseIterator:
    """Iterator for Datapoints Response Objects."""

    def __init__(self, datapoints_objects):
        self.datapoints_objects = datapoints_objects
        self.counter = 0

    def __getitem__(self, index):
        return self.datapoints_objects[index]

    def __len__(self):
        return len(self.datapoints_objects)

    def __iter__(self):
        return self

    def __next__(self):
        if self.counter > len(self.datapoints_objects) - 1:
            raise StopIteration
        else:
            self.counter += 1
            return self.datapoints_objects[self.counter - 1]


class Datapoint(CogniteResource):
    """Data transfer object for datapoints.

    Args:
        timestamp (Union[int, float, datetime]): The data timestamp in milliseconds since the epoch (Jan 1, 1970) or as
            a datetime object.
        value (Union[string, int, float]): The data value, Can be string or numeric depending on the metric.
    """

    def __init__(self, timestamp, value):
        if isinstance(timestamp, datetime):
            self.timestamp = _utils.datetime_to_ms(timestamp)
        else:
            self.timestamp = timestamp
        self.value = value


class TimeseriesWithDatapoints(CogniteResource):
    """Data transfer object for a timeseries with datapoints.

    Args:
        name (str):       Unique ID of time series.
        datapoints (List[stable.datapoints.Datapoint]): List of datapoints in the timeseries.
    """

    def __init__(self, name, datapoints):
        self.name = name
        self.datapoints = datapoints


class LatestDatapointResponse(CogniteResponse):
    """Latest Datapoint Response Object."""

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"][0]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        return pd.DataFrame([self.internal_representation["data"]["items"][0]])


class DatapointsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)
        self._LIMIT_AGG = 10000
        self._LIMIT = 100000

    def get_datapoints(self, name, start, end=None, aggregates=None, granularity=None, **kwargs) -> DatapointsResponse:
        """Returns a DatapointsObject containing a list of datapoints for the given query.

        This method will automate paging for the user and return all data for the given time period.

        Args:
            name (str):             The name of the timeseries to retrieve data for.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step,
                                    continuousvariance/cv, discretevariance/dv, totalvariation/tv'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        Keyword Arguments:
            workers (int):    Number of download workers to run in parallell. Defaults to 10.

            include_outside_points (bool):      No description

            protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.
                                    Defaults to True.

            limit (str):            Max number of datapoints to return. If limit is specified, this method will not automate
                                    paging and will return a maximum of 100,000 dps.

        Returns:
            stable.datapoints.DatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.

        Examples:
            Getting the last 3 days of raw datapoints for a given time series::

                client = CogniteClient()
                res = client.datapoints.get_datapoints(name="my_ts", start="3d-ago")
                print(res.to_pandas())
        """
        start, end = _utils.interval_to_ms(start, end)

        if start >= end:
            raise ValueError("end must be greater than start")

        if aggregates:
            aggregates = ",".join(aggregates)

        if kwargs.get("limit"):
            return self._get_datapoints_user_defined_limit(
                name,
                aggregates,
                granularity,
                start,
                end,
                limit=kwargs.get("limit"),
                protobuf=kwargs.get("protobuf"),
                include_outside_points=kwargs.get("include_outside_points", False),
            )

        num_of_workers = kwargs.get("workers", self._num_of_workers)
        if kwargs.get("include_outside_points") is True:
            num_of_workers = 1

        windows = _utils.get_datapoints_windows(start, end, granularity, num_of_workers)

        partial_get_dps = partial(
            self._get_datapoints_helper_wrapper,
            name=name,
            aggregates=aggregates,
            granularity=granularity,
            protobuf=kwargs.get("protobuf", True),
            include_outside_points=kwargs.get("include_outside_points", False),
        )

        with Pool(len(windows)) as p:
            datapoints = p.map(partial_get_dps, windows)

        concat_dps = []
        [concat_dps.extend(el) for el in datapoints]

        return DatapointsResponse({"data": {"items": [{"name": name, "datapoints": concat_dps}]}})

    def _get_datapoints_helper_wrapper(self, args, name, aggregates, granularity, protobuf, include_outside_points):
        return self._get_datapoints_helper(
            name,
            aggregates,
            granularity,
            args["start"],
            args["end"],
            protobuf=protobuf,
            include_outside_points=include_outside_points,
        )

    def _get_datapoints_helper(self, name, aggregates=None, granularity=None, start=None, end=None, **kwargs):
        """Returns a list of datapoints for the given query.

        This method will automate paging for the given time period.

        Args:
            name (str):       The name of the timeseries to retrieve data for.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step',
                                    continuousvariance/cv, discretevariance/dv, totalvariation/tv'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

        Keyword Arguments:
            include_outside_points (bool):  No description.

            protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.
                                    Defaults to True.

        Returns:
            list of datapoints: A list containing datapoint dicts.
        """
        url = "/timeseries/data/{}".format(quote(name, safe=""))

        use_protobuf = kwargs.get("protobuf", True) and aggregates is None
        limit = self._LIMIT if aggregates is None else self._LIMIT_AGG

        params = {
            "aggregates": aggregates,
            "granularity": granularity,
            "limit": limit,
            "start": start,
            "end": end,
            "includeOutsidePoints": kwargs.get("include_outside_points", False),
        }

        headers = {"accept": "application/protobuf"} if use_protobuf else {}
        datapoints = []
        while (not datapoints or len(datapoints[-1]) == limit) and params["end"] > params["start"]:
            res = self._get(url, params=params, headers=headers)
            if use_protobuf:
                ts_data = _api_timeseries_data_v2_pb2.TimeseriesData()
                ts_data.ParseFromString(res.content)
                parsed_dps = ts_data.numericData.points
                if not parsed_dps:
                    parsed_dps = ts_data.stringData.points
                res = [{"timestamp": p.timestamp, "value": p.value} for p in parsed_dps]
            else:
                res = res.json()["data"]["items"][0]["datapoints"]

            if not res:
                break

            datapoints.append(res)
            latest_timestamp = int(datapoints[-1][-1]["timestamp"])
            params["start"] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
        dps = []
        [dps.extend(el) for el in datapoints]
        return dps

    def _get_datapoints_user_defined_limit(self, name, aggregates, granularity, start, end, limit, **kwargs):
        """Returns a DatapointsResponse object with the requested data.

        No paging or parallelizing is done.

        Args:
            name (str):       The name of the timeseries to retrieve data for.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step',
                                    continuousvariance/cv, discretevariance/dv, totalvariation/tv'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            limit (str):            Max number of datapoints to return. Max is 100,000.

        Keyword Arguments:
            include_outside_points (bool):  No description.

            protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.
                                    Defaults to True.
        Returns:
            stable.datapoints.DatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/timeseries/data/{}".format(quote(name, safe=""))

        use_protobuf = kwargs.get("protobuf", True) and aggregates is None

        params = {
            "aggregates": aggregates,
            "granularity": granularity,
            "limit": limit,
            "start": start,
            "end": end,
            "includeOutsidePoints": kwargs.get("include_outside_points", False),
        }
        headers = {"accept": "application/protobuf"} if use_protobuf else {}
        res = self._get(url, params=params, headers=headers)
        if use_protobuf:
            ts_data = _api_timeseries_data_v2_pb2.TimeseriesData()
            ts_data.ParseFromString(res.content)
            parsed_dps = ts_data.numericData.points
            if not parsed_dps:
                parsed_dps = ts_data.stringData.points
            res = [{"timestamp": p.timestamp, "value": p.value} for p in parsed_dps]
        else:
            res = res.json()["data"]["items"][0]["datapoints"]

        return DatapointsResponse({"data": {"items": [{"name": name, "datapoints": res}]}})

    def _split_TimeseriesWithDatapoints_if_over_limit(
        self, timeseries_with_datapoints: TimeseriesWithDatapoints, limit: int
    ) -> List[TimeseriesWithDatapoints]:
        """Takes a TimeseriesWithDatapoints and splits it into multiple so that each has a max number of datapoints equal
        to the limit given.

        Args:
            timeseries_with_datapoints (stable.datapoints.TimeseriesWithDatapoints): The timeseries with data to potentially split up.

        Returns:
            A list of stable.datapoints.TimeseriesWithDatapoints where each has a maximum number of datapoints equal to the limit given.
        """
        timeseries_with_datapoints_list = []
        if len(timeseries_with_datapoints.datapoints) > limit:
            i = 0
            while i < len(timeseries_with_datapoints.datapoints):
                timeseries_with_datapoints_list.append(
                    TimeseriesWithDatapoints(
                        name=timeseries_with_datapoints.name,
                        datapoints=timeseries_with_datapoints.datapoints[i : i + limit],
                    )
                )
                i += limit
        else:
            timeseries_with_datapoints_list.append(timeseries_with_datapoints)

        return timeseries_with_datapoints_list

    def post_multi_time_series_datapoints(self, timeseries_with_datapoints: Iterable[TimeseriesWithDatapoints]) -> None:
        """Insert data into multiple timeseries.

        Args:
            timeseries_with_datapoints (Iterable[stable.datapoints.TimeseriesWithDatapoints]): The timeseries with data to insert.

        Returns:
            None

        Examples:
            Posting some dummy datapoints to multiple time series. This example assumes that the time series have
            already been created::

                from cognite.client.stable.datapoints import TimeseriesWithDatapoints, Datapoint

                start = 1514761200000
                my_dummy_data_1 = [Datapoint(timestamp=ms, value=i) for i, ms in range(start, start+100)]
                ts_with_datapoints_1 = TimeseriesWithDatapoints(name="ts1", datapoints=my_dummy_data_1)

                start = 1503331800000
                my_dummy_data_2 = [Datapoint(timestamp=ms, value=i) for i, ms in range(start, start+100)]
                ts_with_datapoints_2 = TimeseriesWithDatapoints(name="ts2", datapoints=my_dummy_data_2)

                my_dummy_data = [ts_with_datapoints_1, ts_with_datapoints_2]

                client = CogniteClient()
                client.datapoints.post_multi_time_series_datapoints(my_dummy_data)
        """
        url = "/timeseries/data"

        ul_dps_limit = 100000

        # Make sure we only work with TimeseriesWithDatapoints objects that has a max number of datapoints
        timeseries_with_datapoints_limited = []
        for entry in timeseries_with_datapoints:
            timeseries_with_datapoints_limited.extend(
                self._split_TimeseriesWithDatapoints_if_over_limit(entry, ul_dps_limit)
            )

        # Group these TimeseriesWithDatapoints if possible so that we upload as much as possible in each call to the API
        timeseries_to_upload_binned = _utils.first_fit(
            list_items=timeseries_with_datapoints_limited, max_size=ul_dps_limit, get_count=lambda x: len(x.datapoints)
        )

        for bin in timeseries_to_upload_binned:
            body = {
                "items": [
                    {"name": ts_with_data.name, "datapoints": [dp.__dict__ for dp in ts_with_data.datapoints]}
                    for ts_with_data in bin
                ]
            }
            self._post(url, body=body)

    def post_datapoints(self, name, datapoints: List[Datapoint]) -> None:
        """Insert a list of datapoints.

        Args:
            name (str):       Name of timeseries to insert to.

            datapoints (List[stable.datapoints.Datapoint]): List of datapoint data transfer objects to insert.

        Returns:
            None

        Examples:
            Posting some dummy datapoints::

                from cognite.client.stable.datapoints import Datapoint

                client = CogniteClient()

                start = 1514761200000
                my_dummy_data = [Datapoint(timestamp=start+off, value=off) for off in range(100)]
                client.datapoints.post_datapoints(ts_name, my_dummy_data)
        """
        url = "/timeseries/data/{}".format(quote(name, safe=""))

        ul_dps_limit = 100000
        i = 0
        while i < len(datapoints):
            body = {"items": [dp.__dict__ for dp in datapoints[i : i + ul_dps_limit]]}
            self._post(url, body=body)
            i += ul_dps_limit

    def get_latest(self, name, before=None) -> LatestDatapointResponse:
        """Returns a LatestDatapointObject containing the latest datapoint for the given timeseries.

        Args:
            name (str):       The name of the timeseries to retrieve data for.

        Returns:
            stable.datapoints.LatestDatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.

        Examples:
            Get the latest datapoint from a time series before time x::

                client = CogniteClient()
                x = 1514761200000
                client.datapoints.get_latest(name="my_ts", before=x)

        """
        url = "/timeseries/latest/{}".format(quote(name, safe=""))
        params = {"before": before}
        res = self._get(url, params=params)
        return LatestDatapointResponse(res.json())

    def get_multi_time_series_datapoints(
        self, datapoints_queries: List[DatapointsQuery], start, end=None, aggregates=None, granularity=None, **kwargs
    ) -> DatapointsResponseIterator:
        """Returns a list of DatapointsObjects each of which contains a list of datapoints for the given timeseries.

        This method will automate paging for the user and return all data for the given time period(s).

        Args:
            datapoints_queries (list[stable.datapoints.DatapointsQuery]): The list of DatapointsQuery objects specifying which
                                                                        timeseries to retrieve data for.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            aggregates (list, optional):    The list of aggregate functions you wish to apply to the data. Valid aggregate
                                            functions are: 'average/avg, max, min, count, sum, interpolation/int,
                                            stepinterpolation/step, continuousvariance/cv, discretevariance/dv, totalvariation/tv'.

            granularity (str):              The granularity of the aggregate values. Valid entries are : 'day/d, hour/h,
                                            minute/m, second/s', or a multiple of these indicated by a number as a prefix
                                            e.g. '12hour'.

        Keyword Arguments:
            include_outside_points (bool):  No description.

        Returns:
            stable.datapoints.DatapointsResponseIterator: An iterator which iterates over stable.datapoints.DatapointsResponse objects.
        """
        url = "/timeseries/dataquery"
        start, end = _utils.interval_to_ms(start, end)
        if start >= end:
            raise ValueError("end must be greater than start")

        datapoints_queries = [copy(dpq) for dpq in datapoints_queries]
        num_of_dpqs_with_agg = 0
        num_of_dpqs_raw = 0
        for dpq in datapoints_queries:
            if (dpq.aggregates is None and aggregates is None) or dpq.aggregates == "":
                num_of_dpqs_raw += 1
            else:
                num_of_dpqs_with_agg += 1

        items = []
        for dpq in datapoints_queries:
            if dpq.aggregates is None and aggregates is None:
                dpq.limit = int(self._LIMIT / num_of_dpqs_raw)
            else:
                dpq.limit = int(self._LIMIT_AGG / num_of_dpqs_with_agg)
            items.append(dpq.__dict__)
        body = {
            "items": items,
            "aggregates": ",".join(aggregates) if aggregates is not None else None,
            "granularity": granularity,
            "start": start,
            "includeOutsidePoints": kwargs.get("include_outside_points", False),
            "end": end,
        }
        datapoints_responses = []
        has_incomplete_requests = True
        while has_incomplete_requests:
            res = self._post(url=url, body=body).json()["data"]["items"]
            datapoints_responses.append(res)
            has_incomplete_requests = False
            for i, dpr in enumerate(res):
                dpq = datapoints_queries[i]
                if len(dpr["datapoints"]) == dpq.limit:
                    has_incomplete_requests = True
                    latest_timestamp = dpr["datapoints"][-1]["timestamp"]
                    ts_granularity = granularity if dpq.granularity is None else dpq.granularity
                    next_start = latest_timestamp + (_utils.granularity_to_ms(ts_granularity) if ts_granularity else 1)
                else:
                    next_start = end - 1
                    if datapoints_queries[i].end:
                        next_start = datapoints_queries[i].end - 1
                datapoints_queries[i].start = next_start

        results = [{"data": {"items": [{"name": dpq.name, "datapoints": []}]}} for dpq in datapoints_queries]
        for res in datapoints_responses:
            for i, ts in enumerate(res):
                results[i]["data"]["items"][0]["datapoints"].extend(ts["datapoints"])
        return DatapointsResponseIterator([DatapointsResponse(result) for result in results])

    def get_datapoints_frame(self, time_series, aggregates, granularity, start, end=None, **kwargs) -> pd.DataFrame:
        """Returns a pandas dataframe of datapoints for the given timeseries all on the same timestamps.

        This method will automate paging for the user and return all data for the given time period.

        Args:
            time_series (list):  The list of timeseries names to retrieve data for. Each timeseries can be either a string
                                containing the timeseries or a dictionary containing the names of thetimeseries and a
                                list of specific aggregate functions.

            aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not
                                specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,
                                count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

        Keyword Arguments:
            limit (str): Max number of rows to return. If limit is specified, this method will not automate
                            paging and will return a maximum of 100,000 rows.

            workers (int):    Number of download workers to run in parallell. Defaults to 10.

        Returns:
            pandas.DataFrame: A pandas dataframe containing the datapoints for the given timeseries. The datapoints for all the
            timeseries will all be on the same timestamps.

        Examples:
            Get a dataframe of aggregated time series data::

                client = CogniteClient()

                res = client.datapoints.get_datapoints_frame(time_series=["ts1", "ts2"],
                                aggregates=["avg"], granularity="30s", start="1w-ago")

                print(res)

            The ``timeseries`` parameter can take a list of strings and/or dicts on the following formats.
            This is useful for specifying aggregate functions on a per time series level::

                Using strings:
                    ['<timeseries1>', '<timeseries2>']

                Using dicts:
                    [{'name': '<timeseries1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},
                    {'name': '<timeseries2>', 'aggregates': []}]

                Using both:
                    ['<timeseries1>', {'name': '<timeseries2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]
        """
        if not isinstance(time_series, list):
            raise ValueError("time_series should be a list")
        start, end = _utils.interval_to_ms(start, end)

        if start >= end:
            raise ValueError("end must be greater than start")

        if kwargs.get("limit"):
            return self._get_datapoints_frame_user_defined_limit(
                time_series, aggregates, granularity, start, end, limit=kwargs.get("limit")
            )

        num_of_workers = kwargs.get("workers") or self._num_of_workers

        windows = _utils.get_datapoints_windows(start, end, granularity, num_of_workers)

        partial_get_dpsf = partial(
            self._get_datapoints_frame_helper_wrapper,
            time_series=time_series,
            aggregates=aggregates,
            granularity=granularity,
        )

        with Pool(len(windows)) as p:
            dataframes = p.map(partial_get_dpsf, windows)

        df = pd.concat(dataframes).drop_duplicates(subset="timestamp").reset_index(drop=True)

        return df

    def _get_datapoints_frame_helper_wrapper(self, args, time_series, aggregates, granularity):
        return self._get_datapoints_frame_helper(time_series, aggregates, granularity, args["start"], args["end"])

    def _get_datapoints_frame_helper(self, time_series, aggregates, granularity, start=None, end=None):
        """Returns a pandas dataframe of datapoints for the given timeseries all on the same timestamps.

        This method will automate paging for the user and return all data for the given time period.

        Args:
            time_series (list):     The list of timeseries names to retrieve data for. Each timeseries can be either a string containing the
                                ts name or a dictionary containing the ts name and a list of specific aggregate functions.

            aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not
                                specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,
                                count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

        Returns:
            pandas.DataFrame: A pandas dataframe containing the datapoints for the given timeseries. The datapoints for all the
            timeseries will all be on the same timestamps.

        Note:
            The ``timeseries`` parameter can take a list of strings and/or dicts on the following formats::

                Using strings:
                    ['<timeseries1>', '<timeseries2>']

                Using dicts:
                    [{'name': '<timeseries1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},
                    {'name': '<timeseries2>', 'aggregates': []}]

                Using both:
                    ['<timeseries1>', {'name': '<timeseries2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]
        """
        url = "/timeseries/dataframe"
        num_aggregates = 0
        for ts in time_series:
            if isinstance(ts, str) or ts.get("aggregates") is None:
                num_aggregates += len(aggregates)
            else:
                num_aggregates += len(ts["aggregates"])

        per_tag_limit = int(self._LIMIT / num_aggregates)

        body = {
            "items": [
                {"name": "{}".format(ts)}
                if isinstance(ts, str)
                else {"name": "{}".format(ts["name"]), "aggregates": ts.get("aggregates", [])}
                for ts in time_series
            ],
            "aggregates": aggregates,
            "granularity": granularity,
            "start": start,
            "end": end,
            "limit": per_tag_limit,
        }
        headers = {"accept": "text/csv"}
        dataframes = []
        while (not dataframes or dataframes[-1].shape[0] == per_tag_limit) and body["end"] > body["start"]:
            res = self._post(url=url, body=body, headers=headers)
            dataframes.append(
                pd.read_csv(io.StringIO(res.content.decode(res.encoding if res.encoding else res.apparent_encoding)))
            )
            if dataframes[-1].empty:
                break
            latest_timestamp = int(dataframes[-1].iloc[-1, 0])
            body["start"] = latest_timestamp + _utils.granularity_to_ms(granularity)
        return pd.concat(dataframes).reset_index(drop=True)

    def _get_datapoints_frame_user_defined_limit(self, time_series, aggregates, granularity, start, end, limit):
        """Returns a DatapointsResponse object with the requested data.

        No paging or parallelizing is done.

        Args:
            time_series (List[str]):       The list of timeseries names to retrieve data for. Each timeseries can be either a string containing the
                                ts name or a dictionary containing the ts name and a list of specific aggregate functions.

            aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                    are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

            granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                    second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

            start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                        E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                        epoch or a datetime object which will be converted to ms since epoch UTC.

            end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

            limit (int):            Max number of rows to retrieve. Max is 100,000.

        Returns:
            stable.datapoints.DatapointsResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/timeseries/dataframe"
        body = {
            "items": [
                {"name": "{}".format(ts)}
                if isinstance(ts, str)
                else {"name": "{}".format(ts["name"]), "aggregates": ts.get("aggregates", [])}
                for ts in time_series
            ],
            "aggregates": aggregates,
            "granularity": granularity,
            "start": start,
            "end": end,
            "limit": limit,
        }

        headers = {"accept": "text/csv"}
        res = self._post(url=url, body=body, headers=headers)
        df = pd.read_csv(io.StringIO(res.content.decode(res.encoding if res.encoding else res.apparent_encoding)))

        return df

    def post_datapoints_frame(self, dataframe) -> None:
        """Write a dataframe.
        The dataframe must have a 'timestamp' column with timestamps in milliseconds since epoch.
        The names of the remaining columns specify the names of the time series to which column contents will be written.
        Said time series must already exist.

        Args:
            dataframe (pandas.DataFrame):  Pandas DataFrame Object containing the time series.

        Returns:
            None

        Examples:
            Post a dataframe with white noise::

                client = CogniteClient()
                ts_name = 'NOISE'

                start = datetime(2018, 1, 1)
                # The scaling by 1000 is important: timestamp() returns seconds
                x = [(start + timedelta(days=d)).timestamp() * 1000 for d in range(100)]
                y = np.random.normal(0, 1, 100)

                # The time column must be called precisely 'timestamp'
                df = pd.DataFrame({'timestamp': x, ts_name: y})

                client.datapoints.post_datapoints_frame(df)
        """

        try:
            timestamp = dataframe.timestamp
            names = dataframe.drop(["timestamp"], axis=1).columns
        except:
            raise ValueError("DataFrame not on a correct format")

        for name in names:
            assert not dataframe[name].hasnans, "Dataframe contains NaNs"
            data_points = [Datapoint(int(timestamp.iloc[i]), dataframe[name].iloc[i]) for i in range(0, len(dataframe))]
            self.post_datapoints(name, data_points)

    def live_data_generator(self, name, update_frequency=1):
        """Generator function which continously polls latest datapoint of a timeseries and yields new datapoints.

        Args:
            name (str): Name of timeseries to get latest datapoints for.

            update_frequency (float): Frequency to pull for data in seconds.

        Yields:
            dict: Dictionary containing timestamp and value of latest datapoint.
        """
        last_timestamp = self.get_latest(name).to_json()["timestamp"]
        while True:
            latest = self.get_latest(name).to_json()
            if last_timestamp == latest["timestamp"]:
                time.sleep(update_frequency)
            else:
                yield latest
            last_timestamp = latest["timestamp"]



import json


class APIError(Exception):
    """Cognite API Error

    Raised if a given request fails.

    Args:
        message (str):  The error message produced by the API
        code (int):     The error code produced by the failure
        x_request_id (str): The request-id generated for the failed request.
        extra (Dict):   A dict of any additional information.

    Examples:
        Catching an API-error and handling it based on the error code::

            from cognite.client import CogniteClient, APIError

            client = CogniteClient()

            try:
                client.login.status()
            except APIError as e:
                if e.code == 401:
                    print("You are not authorized")
                elif e.code == 400:
                    print("Something is wrong with your request")
                elif e.code == 500:
                    print("Something went terribly wrong. Here is the request-id: {}".format(e.x_request_id)
                print("The message returned from the API: {}".format(e.message))
    """

    def __init__(self, message, code=None, x_request_id=None, extra=None):
        self.message = message
        self.code = code
        self.x_request_id = x_request_id
        self.extra = extra

    def __str__(self):
        if self.extra:
            pretty_extra = json.dumps(self.extra, indent=4, sort_keys=True)
            return "{} | code: {} | X-Request-ID: {}\n{}".format(
                self.message, self.code, self.x_request_id, pretty_extra
            )
        return "{} | code: {} | X-Request-ID: {}".format(self.message, self.code, self.x_request_id)



# -*- coding: utf-8 -*-
import pandas as pd

from cognite.client._api_client import APIClient, CogniteResponse


class TagMatchingResponse(CogniteResponse):
    """Tag Matching Response Object.

    In addition to the standard output formats this data object also has a to_list() method which returns a list of
    names of the tag matches.
    """

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        matches = []
        for tag in self.internal_representation["data"]["items"]:
            for match in tag["matches"]:
                matches.append(
                    {
                        "tag": tag["tagId"],
                        "match": match["tagId"],
                        "score": match["score"],
                        "platform": match["platform"],
                    }
                )
        if matches:
            return pd.DataFrame(matches)[["tag", "match", "platform", "score"]]
        return pd.DataFrame()

    def to_list(self, first_matches_only=True):
        """Returns a list representation of the matches.

        Args:
            first_matches_only (bool):      Boolean determining whether or not to return only the top match for each
                                            tag.

        Returns:
            list: list of matched tags.
        """
        if self.to_pandas().empty:
            return []
        if first_matches_only:
            return self.to_pandas().sort_values(["score", "match"]).groupby(["tag"]).first()["match"].tolist()
        return self.to_pandas().sort_values(["score", "match"])["match"].tolist()


class TagMatchingClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def tag_matching(self, tag_ids, fuzzy_threshold=0, platform=None) -> TagMatchingResponse:
        """Returns a TagMatchingObject containing a list of matched tags for the given query.

        This method takes an arbitrary string as argument and performs fuzzy matching with a user defined threshold
        toward tag ids in the system.

        Args:
            tag_ids (list):         The tag_ids to retrieve matches for.

            fuzzy_threshold (int):  The threshold to use when searching for matches. A fuzzy threshold of 0 means you only
                                    want to accept perfect matches. Must be >= 0.

            platform (str):         The platform to search on.

        Returns:
            stable.tagmatching.TagMatchingResponse: A data object containing the requested data with several getter methods with different
            output formats.
        """
        url = "/tagmatching"
        body = {"tagIds": tag_ids, "metadata": {"fuzzyThreshold": fuzzy_threshold, "platform": platform}}
        res = self._post(url=url, body=body)
        return TagMatchingResponse(res.json())



# -*- coding: utf-8 -*-
from copy import copy

from cognite.client._api_client import APIClient, CogniteResponse


class LoginStatusResponse(CogniteResponse):
    """
    Attributes:
        user (str): Current user
        logged_in (bool): Is user logged in
        project (str): Current project
        project_id (str): Current project id
    """

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        self.user = internal_representation["data"]["user"]
        self.project = internal_representation["data"]["project"]
        self.project_id = internal_representation["data"]["projectId"]
        self.logged_in = internal_representation["data"]["loggedIn"]

    def to_json(self):
        json_repr = copy(self.__dict__)
        del json_repr["internal_representation"]
        return json_repr


class LoginClient(APIClient):
    def status(self) -> LoginStatusResponse:
        """Check login status

        Returns:
            client.stable.login.LoginStatusResponse

        Examples:
            Check the current login status::

                client = CogniteClient()
                login_status = client.login.status()
                print(login_status)

        """
        return LoginStatusResponse(self._get("/login/status").json())
